@@ -10,7 +10,7 @@
 #include "common.h"
 
 /*
- * MRuby Value definition functions and macros.
+ * mruby Value definition functions and macros.
  */
 MRB_BEGIN_DECL
 
@@ -100,9 +100,13 @@ struct mrb_state;
 # define MRB_ENDIAN_LOHI(a,b) b a
 #endif
 
-MRB_API mrb_int mrb_int_read(const char *p, const char *e, char **endp);
+MRB_API mrb_bool mrb_read_int(const char *p, const char *e, char **endp, mrb_int *np);
+/* obsolete; do not use mrb_int_read() */
+MRB_API mrb_int mrb_int_read(const char*, const char*, char**);
 #ifndef MRB_NO_FLOAT
-MRB_API double mrb_float_read(const char*, char**);
+MRB_API mrb_bool mrb_read_float(const char *p, char **endp, double *fp);
+/* obsolete; do not use mrb_float_read() */
+MRB_API double mrb_float_read(const char *p, char **endp);
 #ifdef MRB_USE_FLOAT32
   typedef float mrb_float;
 #else
@@ -121,7 +125,7 @@ MRB_API int mrb_msvc_snprintf(char *s, size_t n, const char *format, ...);
 #  define isinf(n) (!_finite(n) && !_isnan(n))
 #  define signbit(n) (_copysign(1.0, (n)) < 0.0)
 static const unsigned int IEEE754_INFINITY_BITS_SINGLE = 0x7F800000;
-#  define INFINITY (*(float *)&IEEE754_INFINITY_BITS_SINGLE)
+#  define INFINITY (*(float*)&IEEE754_INFINITY_BITS_SINGLE)
 #  define NAN ((float)(INFINITY - INFINITY))
 # endif
 #endif
@@ -139,22 +143,24 @@ static const unsigned int IEEE754_INFINITY_BITS_SINGLE = 0x7F800000;
   f(MRB_TT_OBJECT,      struct RObject,     "Object") \
   f(MRB_TT_CLASS,       struct RClass,      "Class") \
   f(MRB_TT_MODULE,      struct RClass,      "Module") \
-  f(MRB_TT_ICLASS,      struct RClass,      "iClass") \
   f(MRB_TT_SCLASS,      struct RClass,      "SClass") \
+  f(MRB_TT_HASH,        struct RHash,       "Hash") \
+  f(MRB_TT_CDATA,       struct RData,       "C data") \
+  f(MRB_TT_EXCEPTION,   struct RException,  "Exception") \
+  f(MRB_TT_ICLASS,      struct RClass,      "iClass") \
   f(MRB_TT_PROC,        struct RProc,       "Proc") \
   f(MRB_TT_ARRAY,       struct RArray,      "Array") \
-  f(MRB_TT_HASH,        struct RHash,       "Hash") \
   f(MRB_TT_STRING,      struct RString,     "String") \
   f(MRB_TT_RANGE,       struct RRange,      "Range") \
-  f(MRB_TT_EXCEPTION,   struct RException,  "Exception") \
   f(MRB_TT_ENV,         struct REnv,        "env") \
-  f(MRB_TT_DATA,        struct RData,       "Data") \
   f(MRB_TT_FIBER,       struct RFiber,      "Fiber") \
   f(MRB_TT_STRUCT,      struct RArray,      "Struct") \
   f(MRB_TT_ISTRUCT,     struct RIStruct,    "istruct") \
   f(MRB_TT_BREAK,       struct RBreak,      "break") \
   f(MRB_TT_COMPLEX,     struct RComplex,    "Complex") \
-  f(MRB_TT_RATIONAL,    struct RRational,   "Rational")
+  f(MRB_TT_RATIONAL,    struct RRational,   "Rational") \
+  f(MRB_TT_BIGINT,      struct RBigint,     "Integer") \
+  f(MRB_TT_BACKTRACE,   struct RBacktrace,  "backtrace")
 
 enum mrb_vtype {
 #define MRB_VTYPE_DEFINE(tt, type, name) tt,
@@ -163,6 +169,9 @@ enum mrb_vtype {
   MRB_TT_MAXDEFINE
 };
 
+/* obsolete name for MRB_TT_CDATA */
+#define MRB_TT_DATA MRB_TT_CDATA
+
 #define MRB_VTYPE_TYPEOF(tt) MRB_TYPEOF_##tt
 
 #define MRB_VTYPE_TYPEDEF(tt, type, name) typedef type MRB_VTYPE_TYPEOF(tt);
@@ -178,12 +187,12 @@ MRB_VTYPE_FOREACH(MRB_VTYPE_TYPEDEF)
 
 /**
  * @abstract
- * MRuby value boxing.
+ * mruby value boxing.
  *
  * Actual implementation depends on configured boxing type.
  *
- * @see mruby/boxing_no.h Default boxing representation
- * @see mruby/boxing_word.h Word representation
+ * @see mruby/boxing_word.h Word boxing representation (Default)
+ * @see mruby/boxing_no.h No boxing representation
  * @see mruby/boxing_nan.h Boxed double representation
  */
 typedef void mrb_value;
@@ -237,9 +246,11 @@ struct RCptr {
 #ifndef mrb_true_p
 #define mrb_true_p(o)  (mrb_type(o) == MRB_TT_TRUE)
 #endif
-#ifndef MRB_NO_FLOAT
 #ifndef mrb_float_p
+#ifndef MRB_NO_FLOAT
 #define mrb_float_p(o) (mrb_type(o) == MRB_TT_FLOAT)
+#else
+#define mrb_float_p(o) FALSE
 #endif
 #endif
 #ifndef mrb_array_p
@@ -285,7 +296,7 @@ struct RCptr {
 #define mrb_env_p(o) (mrb_type(o) == MRB_TT_ENV)
 #endif
 #ifndef mrb_data_p
-#define mrb_data_p(o) (mrb_type(o) == MRB_TT_DATA)
+#define mrb_data_p(o) (mrb_type(o) == MRB_TT_CDATA)
 #endif
 #ifndef mrb_fiber_p
 #define mrb_fiber_p(o) (mrb_type(o) == MRB_TT_FIBER)
@@ -300,6 +311,9 @@ struct RCptr {
 #define mrb_bool(o)   (mrb_type(o) != MRB_TT_FALSE)
 #endif
 #define mrb_test(o)   mrb_bool(o)
+#ifndef mrb_bigint_p
+#define mrb_bigint_p(o) (mrb_type(o) == MRB_TT_BIGINT)
+#endif
 
 /**
  * Returns a float in Ruby.
@@ -307,7 +321,8 @@ struct RCptr {
  * Takes a float and boxes it into an mrb_value
  */
 #ifndef MRB_NO_FLOAT
-MRB_INLINE mrb_value mrb_float_value(struct mrb_state *mrb, mrb_float f)
+MRB_INLINE mrb_value
+mrb_float_value(struct mrb_state *mrb, mrb_float f)
 {
   mrb_value v;
   (void) mrb;
@@ -328,14 +343,16 @@ mrb_cptr_value(struct mrb_state *mrb, void *p)
 /**
  * Returns an integer in Ruby.
  */
-MRB_INLINE mrb_value mrb_int_value(struct mrb_state *mrb, mrb_int i)
+MRB_INLINE mrb_value
+mrb_int_value(struct mrb_state *mrb, mrb_int i)
 {
   mrb_value v;
   SET_INT_VALUE(mrb, v, i);
   return v;
 }
 
-MRB_INLINE mrb_value mrb_fixnum_value(mrb_int i)
+MRB_INLINE mrb_value
+mrb_fixnum_value(mrb_int i)
 {
   mrb_value v;
   SET_FIXNUM_VALUE(v, i);
@@ -355,8 +372,6 @@ mrb_obj_value(void *p)
 {
   mrb_value v;
   SET_OBJ_VALUE(v, (struct RBasic*)p);
-  mrb_assert(p == mrb_ptr(v));
-  mrb_assert(((struct RBasic*)p)->tt == mrb_type(v));
   return v;
 }
 
@@ -366,7 +381,8 @@ mrb_obj_value(void *p)
  * @return
  *      nil mrb_value object reference.
  */
-MRB_INLINE mrb_value mrb_nil_value(void)
+MRB_INLINE mrb_value
+mrb_nil_value(void)
 {
   mrb_value v;
   SET_NIL_VALUE(v);
@@ -376,7 +392,8 @@ MRB_INLINE mrb_value mrb_nil_value(void)
 /**
  * Returns false in Ruby.
  */
-MRB_INLINE mrb_value mrb_false_value(void)
+MRB_INLINE mrb_value
+mrb_false_value(void)
 {
   mrb_value v;
   SET_FALSE_VALUE(v);
@@ -386,7 +403,8 @@ MRB_INLINE mrb_value mrb_false_value(void)
 /**
  * Returns true in Ruby.
  */
-MRB_INLINE mrb_value mrb_true_value(void)
+MRB_INLINE mrb_value
+mrb_true_value(void)
 {
   mrb_value v;
   SET_TRUE_VALUE(v);
@@ -424,10 +442,19 @@ mrb_ro_data_p(const char *p)
 #elif defined(__APPLE__)
 #define MRB_LINK_TIME_RO_DATA_P
 #include <mach-o/getsect.h>
+#include <crt_externs.h> // for _NSGetMachExecuteHeader
 static inline mrb_bool
 mrb_ro_data_p(const char *p)
 {
-  return (char*)get_etext() < p && p < (char*)get_edata();
+#ifdef __LP64__
+  struct mach_header_64 *mhp;
+#else
+  struct mach_header *mhp;
+#endif
+  mhp = _NSGetMachExecuteHeader();
+  unsigned long textsize;
+  char *text = (char*)getsegmentdata(mhp, SEG_TEXT, &textsize);
+  return text <= p && p < text + textsize;
 }
 #endif  /* Linux or macOS */
 #endif  /* MRB_NO_DEFAULT_RO_DATA_P */
