@@ -14,12 +14,13 @@
 #include <mruby/opcode.h>
 #include <mruby/variable.h>
 #include <mruby/proc.h>
+#include <mruby/internal.h>
 
 #include "mrdb.h"
 #include "apibreak.h"
 #include "apilist.h"
 
-void mrdb_state_free(mrb_state *);
+void mrdb_state_free(mrb_state*);
 
 static mrb_debug_context *_debug_context = NULL;
 static mrdb_state *_mrdb_state = NULL;
@@ -112,7 +113,7 @@ append_srcpath:
           char *buf;
 
           buflen = strlen(item) + 1;
-          buf = (char *)mrb_malloc(mrb, buflen);
+          buf = (char*)mrb_malloc(mrb, buflen);
           memcpy(buf, item, buflen);
           args->srcpath = buf;
         }
@@ -122,8 +123,7 @@ append_srcpath:
 
           srcpathlen = strlen(args->srcpath);
           itemlen = strlen(item);
-          args->srcpath =
-            (char *)mrb_realloc(mrb, args->srcpath, srcpathlen + itemlen + 2);
+          args->srcpath = (char*)mrb_realloc(mrb, args->srcpath, srcpathlen + itemlen + 2);
           args->srcpath[srcpathlen] = '\n';
           memcpy(args->srcpath + srcpathlen + 1, item, itemlen + 1);
         }
@@ -186,9 +186,9 @@ static mrb_debug_context*
 mrb_debug_context_new(mrb_state *mrb)
 {
   mrb_debug_context *dbg = (mrb_debug_context*)mrb_malloc(mrb, sizeof(mrb_debug_context));
+  static const mrb_debug_context dbg_zero = {0};
 
-  memset(dbg, 0, sizeof(mrb_debug_context));
-
+  *dbg = dbg_zero;
   dbg->xm = DBG_INIT;
   dbg->xphase = DBG_PHASE_BEFORE_RUN;
   dbg->next_bpno = 1;
@@ -225,9 +225,9 @@ static mrdb_state*
 mrdb_state_new(mrb_state *mrb)
 {
   mrdb_state *mrdb = (mrdb_state*)mrb_malloc(mrb, sizeof(mrdb_state));
+  static const mrdb_state mrdb_zero = {0};
 
-  memset(mrdb, 0, sizeof(mrdb_state));
-
+  *mrdb = mrdb_zero;
   mrdb->dbg = mrb_debug_context_get(mrb);
   mrdb->command = (char*)mrb_malloc(mrb, MAX_COMMAND_LINE+1);
   mrdb->print_no = 1;
@@ -279,7 +279,7 @@ get_command(mrb_state *mrb, mrdb_state *mrdb)
   }
 
   if (i == MAX_COMMAND_LINE) {
-    for ( ; (c=getchar()) != EOF && c !='\n'; i++) ;
+    for (; (c=getchar()) != EOF && c !='\n'; i++) ;
   }
 
   if (i > MAX_COMMAND_LINE) {
@@ -296,7 +296,8 @@ pick_out_word(mrb_state *mrb, char **pp)
 {
   char *ps;
 
-  for (ps=*pp; ISBLANK(*ps); ps++) ;
+  for (ps=*pp; ISBLANK(*ps); ps++)
+    ;
   if (*ps == '\0') {
     return NULL;
   }
@@ -335,7 +336,8 @@ parse_command(mrb_state *mrb, mrdb_state *mrdb, char *buf)
   }
   mrdb->wcnt = 1;
   /* set remain parameter */
-  for ( ; *p && ISBLANK(*p); p++) ;
+  for (; *p && ISBLANK(*p); p++)
+    ;
   if (*p) {
     mrdb->words[mrdb->wcnt++] = p;
   }
@@ -355,7 +357,8 @@ parse_command(mrb_state *mrb, mrdb_state *mrdb, char *buf)
       mrdb->words[1] = pick_out_word(mrb, &p);
       if (mrdb->words[1]) {
         /* update remain parameter */
-        for ( ; *p && ISBLANK(*p); p++) ;
+        for (; *p && ISBLANK(*p); p++)
+          ;
         if (*p) {
           mrdb->words[mrdb->wcnt++] = p;
         }
@@ -363,7 +366,7 @@ parse_command(mrb_state *mrb, mrdb_state *mrdb, char *buf)
     }
 
     /* check word #1,#2 */
-    for ( ; cmd->cmd1; cmd++) {
+    for (; cmd->cmd1; cmd++) {
       wlen = strlen(mrdb->words[0]);
       if (wlen < cmd->len1 ||
           strncmp(mrdb->words[0], cmd->cmd1, wlen)) {
@@ -385,7 +388,7 @@ parse_command(mrb_state *mrb, mrdb_state *mrdb, char *buf)
   /* divide remain parameters */
   if (cmd->cmd1 && cmd->div) {
     p = mrdb->words[--mrdb->wcnt];
-    for ( ; mrdb->wcnt<MAX_COMMAND_WORD; mrdb->wcnt++) {
+    for (; mrdb->wcnt<MAX_COMMAND_WORD; mrdb->wcnt++) {
       mrdb->words[mrdb->wcnt] = pick_out_word(mrb, &p);
       if (!mrdb->words[mrdb->wcnt]) {
         break;
@@ -574,7 +577,7 @@ mrb_code_fetch_hook(mrb_state *mrb, const mrb_irep *irep, const mrb_code *pc, mr
 
   switch (dbg->xm) {
   case DBG_STEP:
-    if (!file || (dbg->prvfile == file && dbg->prvline == line)) {
+    if (*pc != OP_JMP && (!file || (dbg->prvfile == file && dbg->prvline == line))) {
       return;
     }
     dbg->method_bpno = 0;
@@ -637,6 +640,7 @@ mrb_code_fetch_hook(mrb_state *mrb, const mrb_irep *irep, const mrb_code *pc, mr
 static mrdb_exemode
 mrb_debug_break_hook(mrb_state *mrb, mrb_debug_context *dbg)
 {
+  ptrdiff_t regs_off = dbg->regs - mrb->c->ci->stack;
   debug_command *cmd;
   dbgcmd_state st = DBGST_CONTINUE;
   mrdb_state *mrdb = mrdb_state_get(mrb);
@@ -648,6 +652,7 @@ mrb_debug_break_hook(mrb_state *mrb, mrb_debug_context *dbg)
     mrb_assert(cmd);
 
     st = cmd->func(mrb, mrdb);
+    dbg->regs = mrb->c->ci->stack + regs_off;
 
     if ((st == DBGST_CONTINUE) || (st == DBGST_RESTART)) break;
   }
@@ -703,10 +708,10 @@ main(int argc, char **argv)
     v = mrb_load_irep_file(mrb, args.rfp);
   }
   else {              /* .rb */
-    mrbc_context *cc = mrbc_context_new(mrb);
-    mrbc_filename(mrb, cc, args.fname);
+    mrb_ccontext *cc = mrb_ccontext_new(mrb);
+    mrb_ccontext_filename(mrb, cc, args.fname);
     v = mrb_load_file_cxt(mrb, args.rfp, cc);
-    mrbc_context_free(mrb, cc);
+    mrb_ccontext_free(mrb, cc);
   }
   if (mrdb->dbg->xm == DBG_QUIT && !mrb_undef_p(v) && mrb->exc) {
     const char *classname = mrb_obj_classname(mrb, mrb_obj_value(mrb->exc));
