@@ -1,9 +1,10 @@
 #include <mruby.h>
+#include <mruby/presym.h>
 
 static mrb_value
 binding_in_c(mrb_state *mrb, mrb_value self)
 {
-  return mrb_funcall_argv(mrb, mrb_obj_value(mrb->object_class), mrb_intern_lit(mrb, "binding"), 0, NULL);
+  return mrb_funcall_argv(mrb, mrb_obj_value(mrb->object_class), MRB_SYM(binding), 0, NULL);
 }
 
 void
