@@ -2,13 +2,14 @@
 ** file.c - File class
 */
 
-#include "mruby.h"
-#include "mruby/class.h"
-#include "mruby/data.h"
-#include "mruby/string.h"
-#include "mruby/ext/io.h"
-#include "mruby/error.h"
-#include "mruby/presym.h"
+#include <mruby.h>
+#include <mruby/array.h>
+#include <mruby/class.h>
+#include <mruby/data.h>
+#include <mruby/string.h>
+#include <mruby/ext/io.h>
+#include <mruby/error.h>
+#include <mruby/presym.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -18,7 +19,7 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
   #include <windows.h>
   #include <io.h>
   #define NULL_FILE "NUL"
@@ -38,19 +39,27 @@
   #define GETCWD getcwd
   #define CHMOD(a, b) chmod(a,b)
   #include <sys/file.h>
+#ifndef __DJGPP__
   #include <libgen.h>
+#endif
   #include <sys/param.h>
   #include <pwd.h>
 #endif
 
 #define FILE_SEPARATOR "/"
 
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
   #define PATH_SEPARATOR ";"
   #define FILE_ALT_SEPARATOR "\\"
   #define VOLUME_SEPARATOR ":"
+  #define DIRSEP_P(ch) (((ch) == '/') | ((ch) == '\\'))
+  #define VOLSEP_P(ch) ((ch) == ':')
+  #define UNC_PATH_P(path) (DIRSEP_P((path)[0]) && DIRSEP_P((path)[1]))
+  #define DRIVE_LETTER_P(path) (((size_t)(((path)[0]) | 0x20) - 'a' <= (size_t)'z' - 'a') && (path)[1] == ':')
+  #define DRIVE_EQUAL_P(x, y) (((x)[0] | 0x20) == ((y)[0] | 0x20))
 #else
   #define PATH_SEPARATOR ":"
+  #define DIRSEP_P(ch) ((ch) == '/')
 #endif
 
 #ifndef LOCK_SH
@@ -82,13 +91,14 @@ typedef struct _stat64      mrb_stat;
 
 #ifdef _WIN32
 static int
-flock(int fd, int operation) {
-  OVERLAPPED ov;
+flock(int fd, int operation)
+{
   HANDLE h = (HANDLE)_get_osfhandle(fd);
   DWORD flags;
   flags = ((operation & LOCK_NB) ? LOCKFILE_FAIL_IMMEDIATELY : 0)
           | ((operation & LOCK_SH) ? LOCKFILE_EXCLUSIVE_LOCK : 0);
-  ov = (OVERLAPPED){0};
+  static const OVERLAPPED zero_ov = {0};
+  OVERLAPPED ov = zero_ov;
   return LockFileEx(h, flags, 0, 0xffffffff, 0xffffffff, &ov) ? 0 : -1;
 }
 #endif
@@ -96,7 +106,7 @@ flock(int fd, int operation) {
 static mrb_value
 mrb_file_s_umask(mrb_state *mrb, mrb_value klass)
 {
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
   /* nothing to do on windows */
   return mrb_fixnum_value(0);
 
@@ -105,7 +115,8 @@ mrb_file_s_umask(mrb_state *mrb, mrb_value klass)
   if (mrb_get_args(mrb, "|i", &mask) == 0) {
     omask = umask(0);
     umask(omask);
-  } else {
+  }
+  else {
     omask = umask(mask);
   }
   return mrb_fixnum_value(omask);
@@ -116,16 +127,14 @@ static mrb_value
 mrb_file_s_unlink(mrb_state *mrb, mrb_value obj)
 {
   const mrb_value *argv;
-  mrb_int argc, i;
-  char *path;
+  mrb_int argc;
 
   mrb_get_args(mrb, "*", &argv, &argc);
-  for (i = 0; i < argc; i++) {
-    const char *utf8_path;
+  for (int i = 0; i < argc; i++) {
     mrb_value pathv = argv[i];
     mrb_ensure_string_type(mrb, pathv);
-    utf8_path = RSTRING_CSTR(mrb, pathv);
-    path = mrb_locale_from_utf8(utf8_path, -1);
+    const char *utf8_path = RSTRING_CSTR(mrb, pathv);
+    char *path = mrb_locale_from_utf8(utf8_path, -1);
     if (UNLINK(path) < 0) {
       mrb_locale_free(path);
       mrb_sys_fail(mrb, utf8_path);
@@ -139,13 +148,12 @@ static mrb_value
 mrb_file_s_rename(mrb_state *mrb, mrb_value obj)
 {
   mrb_value from, to;
-  char *src, *dst;
 
   mrb_get_args(mrb, "SS", &from, &to);
-  src = mrb_locale_from_utf8(RSTRING_CSTR(mrb, from), -1);
-  dst = mrb_locale_from_utf8(RSTRING_CSTR(mrb, to), -1);
+  char *src = mrb_locale_from_utf8(RSTRING_CSTR(mrb, from), -1);
+  char *dst = mrb_locale_from_utf8(RSTRING_CSTR(mrb, to), -1);
   if (rename(src, dst) < 0) {
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
     if (CHMOD(dst, 0666) == 0 && UNLINK(dst) == 0 && rename(src, dst) == 0) {
       mrb_locale_free(src);
       mrb_locale_free(dst);
@@ -162,61 +170,87 @@ mrb_file_s_rename(mrb_state *mrb, mrb_value obj)
   return mrb_fixnum_value(0);
 }
 
+#define SKIP_DIRSEP(p) for (; DIRSEP_P(*(p)); (p)++)
+#define NEXT_DIRSEP(p) for (; *(p) != '\0' && !DIRSEP_P(*(p)); (p)++)
+
+static const char*
+scan_dirname(const char *path, mrb_int level)
+{
+  const char *p = path + strlen(path);
+  if (level < 1) return p;
+  for (; p > path && DIRSEP_P(p[-1]); p--)
+    ;
+  for (; level > 0; level--) {
+    for (; p > path && !DIRSEP_P(p[-1]); p--)
+      ;
+    for (; p > path && DIRSEP_P(p[-1]); p--)
+      ;
+  }
+  return p > path ? p : path;
+}
+
 static mrb_value
 mrb_file_dirname(mrb_state *mrb, mrb_value klass)
 {
-#if defined(_WIN32) || defined(_WIN64)
-  char dname[_MAX_DIR], vname[_MAX_DRIVE];
-  char buffer[_MAX_DRIVE + _MAX_DIR];
-  const char *utf8_path;
-  char *path;
-  size_t ridx;
-  mrb_get_args(mrb, "z", &utf8_path);
-  path = mrb_locale_from_utf8(utf8_path, -1);
-  _splitpath(path, vname, dname, NULL, NULL);
-  snprintf(buffer, _MAX_DRIVE + _MAX_DIR, "%s%s", vname, dname);
-  mrb_locale_free(path);
-  ridx = strlen(buffer);
-  if (ridx == 0) {
-    strncpy(buffer, ".", 2);  /* null terminated */
-  } else if (ridx > 1) {
-    ridx--;
-    while (ridx > 0 && (buffer[ridx] == '/' || buffer[ridx] == '\\')) {
-      buffer[ridx] = '\0';  /* remove last char */
-      ridx--;
-    }
-  }
-  return mrb_str_new_cstr(mrb, buffer);
-#else
-  char *dname, *path;
-  mrb_value s;
-  mrb_get_args(mrb, "S", &s);
-  path = mrb_locale_from_utf8(mrb_str_to_cstr(mrb, s), -1);
+  const char *path;
+  mrb_int level = 1;
+  mrb_get_args(mrb, "z|i", &path, &level);
 
-  if ((dname = dirname(path)) == NULL) {
-    mrb_locale_free(path);
-    mrb_sys_fail(mrb, "dirname");
+  if (level < 0) {
+    mrb_raisef(mrb, E_ARGUMENT_ERROR, "negative level: %i", level);
+  }
+
+  const char *p = path;
+#ifdef _WIN32
+  if (UNC_PATH_P(p)) {
+    p += 2;
+    SKIP_DIRSEP(p);
+    path = p - 2; /* if consecutive, point to the trailing slash */
+    NEXT_DIRSEP(p);
+    const char *o = p;
+    SKIP_DIRSEP(p);
+    if (*p == '\0') {
+      p = o;
+    }
+    else {
+      NEXT_DIRSEP(p);
+      p = scan_dirname(p, level);
+    }
+    return mrb_str_new(mrb, path, p - path);
+  }
+  else if (ISALPHA(p[0]) && p[1] == ':') {
+    p += 2;
+    const char *o = p;
+    SKIP_DIRSEP(p);
+    p = scan_dirname(p, level);
+    mrb_value s = mrb_str_new(mrb, path, p - path);
+    if (p == o) {
+      mrb_str_cat_lit(mrb, s, ".");
+    }
+    return s;
   }
-  mrb_locale_free(path);
-  return mrb_str_new_cstr(mrb, dname);
 #endif
+  SKIP_DIRSEP(p);
+  if (p > path) {
+    path = p - 1; /* if consecutive, point to the trailing slash */
+  }
+  p = scan_dirname(p, level);
+  return (p == path) ? mrb_str_new_lit(mrb, ".") : mrb_str_new(mrb, path, p - path);
 }
 
 static mrb_value
 mrb_file_basename(mrb_state *mrb, mrb_value klass)
 {
   // NOTE: Do not use mrb_locale_from_utf8 here
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
   char bname[_MAX_DIR];
   char extname[_MAX_EXT];
-  char *path;
-  size_t ridx;
   char buffer[_MAX_DIR + _MAX_EXT];
   mrb_value s;
 
   mrb_get_args(mrb, "S", &s);
-  path = mrb_str_to_cstr(mrb, s);
-  ridx = strlen(path);
+  char *path = mrb_str_to_cstr(mrb, s);
+  size_t ridx = strlen(path);
   if (ridx > 0) {
     ridx--;
     while (ridx > 0 && (path[ridx] == '/' || path[ridx] == '\\')) {
@@ -231,10 +265,10 @@ mrb_file_basename(mrb_state *mrb, mrb_value klass)
   snprintf(buffer, _MAX_DIR + _MAX_EXT, "%s%s", bname, extname);
   return mrb_str_new_cstr(mrb, buffer);
 #else
-  char *bname, *path;
   mrb_value s;
   mrb_get_args(mrb, "S", &s);
-  path = mrb_str_to_cstr(mrb, s);
+  char *path = mrb_str_to_cstr(mrb, s);
+  char *bname;
   if ((bname = basename(path)) == NULL) {
     mrb_sys_fail(mrb, "basename");
   }
@@ -246,19 +280,16 @@ mrb_file_basename(mrb_state *mrb, mrb_value klass)
 static mrb_value
 mrb_file_realpath(mrb_state *mrb, mrb_value klass)
 {
-  mrb_value pathname, dir_string, s, result;
-  mrb_int argc;
-  char *cpath;
-
-  argc = mrb_get_args(mrb, "S|S", &pathname, &dir_string);
+  mrb_value pathname, dir_string;
+  mrb_int argc = mrb_get_args(mrb, "S|S", &pathname, &dir_string);
   if (argc == 2) {
-    s = mrb_str_dup(mrb, dir_string);
+    mrb_value s = mrb_str_dup(mrb, dir_string);
     s = mrb_str_append(mrb, s, mrb_str_new_cstr(mrb, FILE_SEPARATOR));
     s = mrb_str_append(mrb, s, pathname);
     pathname = s;
   }
-  cpath = mrb_locale_from_utf8(RSTRING_CSTR(mrb, pathname), -1);
-  result = mrb_str_new_capa(mrb, PATH_MAX);
+  char *cpath = mrb_locale_from_utf8(RSTRING_CSTR(mrb, pathname), -1);
+  mrb_value result = mrb_str_new_capa(mrb, PATH_MAX);
   if (realpath(cpath, RSTRING_PTR(result)) == NULL) {
     mrb_locale_free(cpath);
     mrb_sys_fail(mrb, RSTRING_CSTR(mrb, pathname));
@@ -269,126 +300,303 @@ mrb_file_realpath(mrb_state *mrb, mrb_value klass)
   return result;
 }
 
-static mrb_value
-mrb_file__getwd(mrb_state *mrb, mrb_value klass)
+static const char*
+path_getwd(mrb_state *mrb)
 {
-  mrb_value path;
-  char buf[MAXPATHLEN], *utf8;
+  char buf[MAXPATHLEN];
 
   if (GETCWD(buf, MAXPATHLEN) == NULL) {
     mrb_sys_fail(mrb, "getcwd(2)");
   }
-  utf8 = mrb_utf8_from_locale(buf, -1);
-  path = mrb_str_new_cstr(mrb, utf8);
+  char *utf8 = mrb_utf8_from_locale(buf, -1);
+  mrb_value path = mrb_str_new_cstr(mrb, utf8);
   mrb_utf8_free(utf8);
-  return path;
+  return RSTRING_CSTR(mrb, path);
 }
 
+static mrb_bool
+path_absolute_p(const char *path)
+{
 #ifdef _WIN32
-#define IS_FILESEP(x) (x == (*(char*)(FILE_SEPARATOR)) || x == (*(char*)(FILE_ALT_SEPARATOR)))
-#define IS_VOLSEP(x) (x == (*(char*)(VOLUME_SEPARATOR)))
-#define IS_DEVICEID(x) (x == '.' || x == '?')
-#define CHECK_UNCDEV_PATH (IS_FILESEP(path[0]) && IS_FILESEP(path[1]))
-
-static int
-is_absolute_traditional_path(const char *path, size_t len)
-{
-  if (len < 3) return 0;
-  return (ISALPHA(path[0]) && IS_VOLSEP(path[1]) && IS_FILESEP(path[2]));
-}
-
-static int
-is_absolute_unc_path(const char *path, size_t len) {
-  if (len < 2) return 0;
-  return (CHECK_UNCDEV_PATH && !IS_DEVICEID(path[2]));
-}
-
-static int
-is_absolute_device_path(const char *path, size_t len) {
-  if (len < 4) return 0;
-  return (CHECK_UNCDEV_PATH && IS_DEVICEID(path[2]) && IS_FILESEP(path[3]));
-}
-
-static int
-mrb_file_is_absolute_path(const char *path)
-{
-  size_t len = strlen(path);
-  if (IS_FILESEP(path[0])) return 1;
-  if (len > 0)
-    return (
-      is_absolute_traditional_path(path, len) ||
-      is_absolute_unc_path(path, len) ||
-      is_absolute_device_path(path, len)
-      );
-  else
-    return 0;
-}
-
-#undef IS_FILESEP
-#undef IS_VOLSEP
-#undef IS_DEVICEID
-#undef CHECK_UNCDEV_PATH
-
+  return UNC_PATH_P(path) ||
+         (ISALPHA(path[0]) && VOLSEP_P(path[1]) && DIRSEP_P(path[2]));
 #else
-static int
-mrb_file_is_absolute_path(const char *path)
-{
-  return (path[0] == *(char*)(FILE_SEPARATOR));
-}
+  return DIRSEP_P(path[0]);
 #endif
+}
 
-static mrb_value
-mrb_file__gethome(mrb_state *mrb, mrb_value klass)
+static void
+path_parse(mrb_state *mrb, mrb_value ary, const char *path, int ai)
 {
-  mrb_int argc;
+#ifdef _WIN32
+  if (DRIVE_LETTER_P(path)) {
+    mrb_ary_set(mrb, ary, 0, mrb_str_new(mrb, path, 2));
+    path += 2;
+    if (DIRSEP_P(*path)) {
+      ARY_SET_LEN(mrb_ary_ptr(ary), 1);
+    }
+    mrb_gc_arena_restore(mrb, ai);
+  }
+  else if (UNC_PATH_P(path)) {
+    path += 2;
+    SKIP_DIRSEP(path);
+    const char *path0 = path;
+    NEXT_DIRSEP(path);
+    mrb_value prefix = mrb_str_new_lit(mrb, "//");
+    mrb_str_cat(mrb, prefix, path0, path - path0);
+    ARY_SET_LEN(mrb_ary_ptr(ary), 0);
+    mrb_ary_push(mrb, ary, prefix);
+    mrb_gc_arena_restore(mrb, ai);
+  }
+  else
+#endif /* _WIN32 */
+  {
+    if (RARRAY_LEN(ary) == 0) {
+      mrb_ary_set(mrb, ary, 0, mrb_nil_value());
+    }
+    else if (DIRSEP_P(*path)) {
+      ARY_SET_LEN(mrb_ary_ptr(ary), 1);
+    }
+  }
+
+  for (;;) {
+    SKIP_DIRSEP(path);
+    const char *path0 = path;
+    NEXT_DIRSEP(path);
+    int len = path - path0;
+    if (len == 0) {
+      break;
+    }
+    else if (len == 1 && path0[0] == '.') {
+      /* do nothing */
+    }
+    else if (len == 2 && path0[0] == '.' && path0[1] == '.') {
+      if (RARRAY_LEN(ary) >= 2) {
+        mrb_ary_pop(mrb, ary);
+      }
+    }
+    else {
+      mrb_ary_push(mrb, ary, mrb_str_new(mrb, path0, path - path0));
+      mrb_gc_arena_restore(mrb, ai);
+    }
+  }
+}
+
+// This function decomposes path into an array based on basedir and workdir.
+// The array consists of the root prefix at ary[0], zero or more directories, and finally file names.
+// The root prefix is nil for non-Windows, or the drive name or UNC host name for Windows.
+static mrb_value
+path_split(mrb_state *mrb, const char *path, const char *basedir, const char *workdir)
+{
+  mrb_value ary = mrb_ary_new(mrb);
+  int ai = mrb_gc_arena_save(mrb);
+
+  if (workdir) {
+    path_parse(mrb, ary, workdir, ai);
+  }
+
+  if (basedir) {
+    path_parse(mrb, ary, basedir, ai);
+  }
+
+  path_parse(mrb, ary, path, ai);
+
+  return ary;
+}
+
+static const char*
+path_gethome(mrb_state *mrb, const char **pathp)
+{
+  mrb_assert(pathp && *pathp && **pathp == '~');
+
   char *home;
   mrb_value path;
 
-#ifndef _WIN32
-  mrb_value username;
+  const char *username = ++*pathp;
+  NEXT_DIRSEP(*pathp);
+  int len = *pathp - username;
 
-  argc = mrb_get_args(mrb, "|S", &username);
-  if (argc == 0) {
+  if (len == 0) {
     home = getenv("HOME");
+#ifdef _WIN32
     if (home == NULL) {
-      return mrb_nil_value();
+      home = getenv("USERPROFILE");
     }
-    if (!mrb_file_is_absolute_path(home)) {
+#endif
+    if (home == NULL) {
+      mrb_raise(mrb, E_ARGUMENT_ERROR, "couldn't find HOME environment -- expanding '~'");
+    }
+    if (!path_absolute_p(home)) {
       mrb_raise(mrb, E_ARGUMENT_ERROR, "non-absolute home");
     }
-  } else {
-    const char *cuser = RSTRING_CSTR(mrb, username);
-    struct passwd *pwd = getpwnam(cuser);
+  }
+  else {
+    const char *uname = RSTRING_CSTR(mrb, mrb_str_new(mrb, username, len));
+#if defined(_WIN32) || defined(MRB_IO_NO_PWNAM)
+    mrb_raisef(mrb, E_ARGUMENT_ERROR, "user %s doesn't exist", uname);
+#else
+    const struct passwd *pwd = getpwnam(uname);
     if (pwd == NULL) {
-      return mrb_nil_value();
+      mrb_raisef(mrb, E_ARGUMENT_ERROR, "user %s doesn't exist", uname);
     }
     home = pwd->pw_dir;
-    if (!mrb_file_is_absolute_path(home)) {
-      mrb_raisef(mrb, E_ARGUMENT_ERROR, "non-absolute home of ~%v", username);
+    if (!path_absolute_p(home)) {
+      mrb_raisef(mrb, E_ARGUMENT_ERROR, "non-absolute home of ~%s", uname);
     }
-  }
-  home = mrb_locale_from_utf8(home, -1);
-  path = mrb_str_new_cstr(mrb, home);
-  mrb_locale_free(home);
-  return path;
-#else  /* _WIN32 */
-  argc = mrb_get_argc(mrb);
-  if (argc == 0) {
-    home = getenv("USERPROFILE");
-    if (home == NULL) {
-      return mrb_nil_value();
-    }
-    if (!mrb_file_is_absolute_path(home)) {
-      mrb_raise(mrb, E_ARGUMENT_ERROR, "non-absolute home");
-    }
-  } else {
-    return mrb_nil_value();
-  }
-  home = mrb_locale_from_utf8(home, -1);
-  path = mrb_str_new_cstr(mrb, home);
-  mrb_locale_free(home);
-  return path;
 #endif
+  }
+  home = mrb_utf8_from_locale(home, -1);
+  path = mrb_str_new_cstr(mrb, home);
+  mrb_utf8_free(home);
+
+  SKIP_DIRSEP(*pathp);
+  return RSTRING_CSTR(mrb, path);
+}
+
+static mrb_value
+path_expand(mrb_state *mrb, const char *path, const char *base, mrb_bool tilda)
+{
+  mrb_value ary;
+
+  // split path conponents as array and normalization
+  if (tilda && path[0] == '~') {
+    base = path_gethome(mrb, &path);
+    ary = path_split(mrb, path, base, NULL);
+  }
+  else if (path_absolute_p(path)) {
+    ary = path_split(mrb, path, NULL, NULL);
+  }
+  else {
+    const char *wd = NULL;
+    if (tilda && base[0] == '~') {
+      wd = path_gethome(mrb, &base);
+    }
+#ifndef _WIN32
+    else if (!path_absolute_p(base)) {
+      wd = path_getwd(mrb);
+    }
+#else
+    else if (DRIVE_LETTER_P(path)) {
+      if (DRIVE_LETTER_P(base) && DRIVE_EQUAL_P(path, base) && DIRSEP_P(base[2])) {
+        wd = NULL;
+      }
+      else {
+        wd = path_getwd(mrb);
+        if (UNC_PATH_P(base) || (DRIVE_LETTER_P(base) && !DRIVE_EQUAL_P(path, base))) {
+          base = NULL;
+        }
+        if (!DRIVE_EQUAL_P(path, wd)) {
+          wd = NULL;
+        }
+      }
+    }
+    else if (UNC_PATH_P(base)) {
+      wd = NULL;
+    }
+    else {
+      wd = path_getwd(mrb);
+    }
+#endif /* _WIN32 */
+    ary = path_split(mrb, path, base, wd);
+  }
+
+  // join path components as string
+  mrb_value ret;
+  mrb_assert(RARRAY_LEN(ary) >= 1);
+#ifndef _WIN32
+  mrb_assert(mrb_nil_p(RARRAY_PTR(ary)[0]));
+  ret = mrb_str_new(mrb, NULL, 0);
+#else
+  mrb_assert(mrb_string_p(RARRAY_PTR(ary)[0]));
+  ret = RARRAY_PTR(ary)[0];
+#endif
+  if (RARRAY_LEN(ary) == 1) {
+#ifdef _WIN32
+    mrb_assert(mrb_string_p(ret));
+    mrb_assert(RSTRING_LEN(ret) >= 2); // drive letter or UNC prefix
+    if (!DIRSEP_P(RSTRING_PTR(ret)[0]))
+#endif
+    {
+      mrb_str_cat_lit(mrb, ret, "/");
+    }
+  }
+  else {
+    for (int i = 1; i < RARRAY_LEN(ary); i++) {
+      mrb_str_cat_lit(mrb, ret, "/");
+      mrb_assert(mrb_string_p(RARRAY_PTR(ary)[i]));
+      mrb_str_cat_str(mrb, ret, RARRAY_PTR(ary)[i]);
+    }
+  }
+
+  return ret;
+}
+
+static mrb_value
+mrb_file_expand_path(mrb_state *mrb, mrb_value self)
+{
+  const char *path;
+  const char *default_dir = ".";
+  mrb_get_args(mrb, "z|z", &path, &default_dir);
+  return path_expand(mrb, path, default_dir, TRUE);
+}
+
+static mrb_value
+mrb_file_absolute_path(mrb_state *mrb, mrb_value self)
+{
+  const char *path;
+  const char *default_dir = ".";
+  mrb_get_args(mrb, "z|z", &path, &default_dir);
+  return path_expand(mrb, path, default_dir, FALSE);
+}
+
+static mrb_value
+mrb_file_absolute_path_p(mrb_state *mrb, mrb_value klass)
+{
+  mrb_value path = mrb_get_arg1(mrb);
+  mrb_ensure_string_type(mrb, path);
+  return mrb_bool_value(path_absolute_p(RSTRING_CSTR(mrb, path)));
+}
+
+#define TIME_OVERFLOW_P(a) (sizeof(time_t) >= sizeof(mrb_int) && ((a) > MRB_INT_MAX || (a) < MRB_INT_MIN))
+#define TIME_T_UINT (~(time_t)0 > 0)
+#if defined(MRB_USE_BITINT)
+#define TIME_BIGTIME(mrb, a)                                                   \
+  return (TIME_T_UINT ? mrb_bint_new_uint64((mrb), (uint64_t)(a))              \
+               : mrb_bint_new_int64(mrb, (int64_t)(a)))
+#elif !defined(MRB_NO_FLOAT)
+#define TIME_BIGTIME(mrb,a) return mrb_float_value((mrb), (mrb_float)(a))
+#else
+#define TIME_BIGTIME(mrb, a) mrb_raise(mrb, E_IO_ERROR, #a " overflow")
+#endif
+
+static mrb_value
+mrb_file_atime(mrb_state *mrb, mrb_value self)
+{
+  int fd = mrb_io_fileno(mrb, self);
+  mrb_stat st;
+
+  mrb->c->ci->mid = 0;
+  if (mrb_fstat(fd, &st) == -1)
+    mrb_sys_fail(mrb, "atime");
+  if (TIME_OVERFLOW_P(st.st_atime)) {
+    TIME_BIGTIME(mrb, st.st_atime);
+  }
+  return mrb_int_value(mrb, (mrb_int)st.st_atime);
+}
+
+static mrb_value
+mrb_file_ctime(mrb_state *mrb, mrb_value self)
+{
+  int fd = mrb_io_fileno(mrb, self);
+  mrb_stat st;
+
+  mrb->c->ci->mid = 0;
+  if (mrb_fstat(fd, &st) == -1)
+    mrb_sys_fail(mrb, "ctime");
+  if (TIME_OVERFLOW_P(st.st_ctime)) {
+    TIME_BIGTIME(mrb, st. st_ctime);
+  }
+  return mrb_int_value(mrb, (mrb_int)st.st_ctime);
 }
 
 static mrb_value
@@ -397,8 +605,12 @@ mrb_file_mtime(mrb_state *mrb, mrb_value self)
   int fd = mrb_io_fileno(mrb, self);
   mrb_stat st;
 
+  mrb->c->ci->mid = 0;
   if (mrb_fstat(fd, &st) == -1)
-    return mrb_false_value();
+    mrb_sys_fail(mrb, "mtime");
+  if (TIME_OVERFLOW_P(st.st_mtime)) {
+    TIME_BIGTIME(mrb, st. st_mtime);
+  }
   return mrb_int_value(mrb, (mrb_int)st.st_mtime);
 }
 
@@ -409,10 +621,9 @@ mrb_file_flock(mrb_state *mrb, mrb_value self)
   mrb_raise(mrb, E_NOTIMP_ERROR, "flock is not supported on Illumos/Solaris/Windows");
 #else
   mrb_int operation;
-  int fd;
 
   mrb_get_args(mrb, "i", &operation);
-  fd = mrb_io_fileno(mrb, self);
+  int fd = mrb_io_fileno(mrb, self);
 
   while (flock(fd, (int)operation) == -1) {
     switch (errno) {
@@ -440,9 +651,7 @@ static mrb_value
 mrb_file_size(mrb_state *mrb, mrb_value self)
 {
   mrb_stat st;
-  int fd;
-
-  fd = mrb_io_fileno(mrb, self);
+  int fd = mrb_io_fileno(mrb, self);
   if (mrb_fstat(fd, &st) == -1) {
     mrb_raise(mrb, E_RUNTIME_ERROR, "fstat failed");
   }
@@ -464,10 +673,8 @@ mrb_ftruncate(int fd, mrb_int length)
 #ifndef _WIN32
   return ftruncate(fd, (off_t)length);
 #else
-  HANDLE file;
   __int64 cur;
-
-  file = (HANDLE)_get_osfhandle(fd);
+  HANDLE file = (HANDLE)_get_osfhandle(fd);
   if (file == INVALID_HANDLE_VALUE) {
     return -1;
   }
@@ -491,12 +698,9 @@ mrb_ftruncate(int fd, mrb_int length)
 static mrb_value
 mrb_file_truncate(mrb_state *mrb, mrb_value self)
 {
-  int fd;
-  mrb_int length;
   mrb_value lenv = mrb_get_arg1(mrb);
-
-  fd = mrb_io_fileno(mrb, self);
-  length = mrb_as_int(mrb, lenv);
+  int fd = mrb_io_fileno(mrb, self);
+  mrb_int length = mrb_as_int(mrb, lenv);
   if (mrb_ftruncate(fd, length) != 0) {
     mrb_raise(mrb, E_IO_ERROR, "ftruncate failed");
   }
@@ -507,16 +711,14 @@ mrb_file_truncate(mrb_state *mrb, mrb_value self)
 static mrb_value
 mrb_file_s_symlink(mrb_state *mrb, mrb_value klass)
 {
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
   mrb_raise(mrb, E_NOTIMP_ERROR, "symlink is not supported on this platform");
 #else
   mrb_value from, to;
-  const char *src, *dst;
-  int ai = mrb_gc_arena_save(mrb);
 
   mrb_get_args(mrb, "SS", &from, &to);
-  src = mrb_locale_from_utf8(RSTRING_CSTR(mrb, from), -1);
-  dst = mrb_locale_from_utf8(RSTRING_CSTR(mrb, to), -1);
+  const char *src = mrb_locale_from_utf8(RSTRING_CSTR(mrb, from), -1);
+  const char *dst = mrb_locale_from_utf8(RSTRING_CSTR(mrb, to), -1);
   if (symlink(src, dst) == -1) {
     mrb_locale_free(src);
     mrb_locale_free(dst);
@@ -524,20 +726,20 @@ mrb_file_s_symlink(mrb_state *mrb, mrb_value klass)
   }
   mrb_locale_free(src);
   mrb_locale_free(dst);
-  mrb_gc_arena_restore(mrb, ai);
 #endif
   return mrb_fixnum_value(0);
 }
 
 static mrb_value
-mrb_file_s_chmod(mrb_state *mrb, mrb_value klass) {
+mrb_file_s_chmod(mrb_state *mrb, mrb_value klass)
+{
   mrb_int mode;
-  mrb_int argc, i;
+  mrb_int argc;
   const mrb_value *filenames;
   int ai = mrb_gc_arena_save(mrb);
 
   mrb_get_args(mrb, "i*", &mode, &filenames, &argc);
-  for (i = 0; i < argc; i++) {
+  for (int i = 0; i < argc; i++) {
     const char *utf8_path = RSTRING_CSTR(mrb, filenames[i]);
     char *path = mrb_locale_from_utf8(utf8_path, -1);
     if (CHMOD(path, mode) == -1) {
@@ -552,8 +754,9 @@ mrb_file_s_chmod(mrb_state *mrb, mrb_value klass) {
 }
 
 static mrb_value
-mrb_file_s_readlink(mrb_state *mrb, mrb_value klass) {
-#if defined(_WIN32) || defined(_WIN64)
+mrb_file_s_readlink(mrb_state *mrb, mrb_value klass)
+{
+#if defined(_WIN32)
   mrb_raise(mrb, E_NOTIMP_ERROR, "readlink is not supported on this platform");
   return mrb_nil_value(); // unreachable
 #else
@@ -562,15 +765,14 @@ mrb_file_s_readlink(mrb_state *mrb, mrb_value klass) {
   size_t bufsize = 100;
   ssize_t rc;
   mrb_value ret;
-  int ai = mrb_gc_arena_save(mrb);
 
   mrb_get_args(mrb, "z", &path);
   tmp = mrb_locale_from_utf8(path, -1);
 
-  buf = (char *)mrb_malloc(mrb, bufsize);
+  buf = (char*)mrb_malloc(mrb, bufsize);
   while ((rc = readlink(tmp, buf, bufsize)) == (ssize_t)bufsize && rc != -1) {
     bufsize *= 2;
-    buf = (char *)mrb_realloc(mrb, buf, bufsize);
+    buf = (char*)mrb_realloc(mrb, buf, bufsize);
   }
   mrb_locale_free(tmp);
   if (rc == -1) {
@@ -579,10 +781,9 @@ mrb_file_s_readlink(mrb_state *mrb, mrb_value klass) {
   }
   tmp = mrb_utf8_from_locale(buf, -1);
   ret = mrb_str_new(mrb, tmp, rc);
-  mrb_locale_free(tmp);
+  mrb_utf8_free(tmp);
   mrb_free(mrb, buf);
 
-  mrb_gc_arena_restore(mrb, ai);
   return ret;
 #endif
 }
@@ -590,38 +791,39 @@ mrb_file_s_readlink(mrb_state *mrb, mrb_value klass) {
 void
 mrb_init_file(mrb_state *mrb)
 {
-  struct RClass *io, *file, *cnst;
-
-  io   = mrb_class_get_id(mrb, MRB_SYM(IO));
-  file = mrb_define_class(mrb, "File", io);
+  struct RClass *io   = mrb_class_get_id(mrb, MRB_SYM(IO));
+  struct RClass *file = mrb_define_class_id(mrb, MRB_SYM(File), io);
   MRB_SET_INSTANCE_TT(file, MRB_TT_CDATA);
-  mrb_define_class_method(mrb, file, "umask",  mrb_file_s_umask, MRB_ARGS_OPT(1));
-  mrb_define_class_method(mrb, file, "delete", mrb_file_s_unlink, MRB_ARGS_ANY());
-  mrb_define_class_method(mrb, file, "unlink", mrb_file_s_unlink, MRB_ARGS_ANY());
-  mrb_define_class_method(mrb, file, "rename", mrb_file_s_rename, MRB_ARGS_REQ(2));
-  mrb_define_class_method(mrb, file, "symlink", mrb_file_s_symlink, MRB_ARGS_REQ(2));
-  mrb_define_class_method(mrb, file, "chmod", mrb_file_s_chmod, MRB_ARGS_REQ(1) | MRB_ARGS_REST());
-  mrb_define_class_method(mrb, file, "readlink", mrb_file_s_readlink, MRB_ARGS_REQ(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(umask),  mrb_file_s_umask, MRB_ARGS_OPT(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(delete), mrb_file_s_unlink, MRB_ARGS_ANY());
+  mrb_define_class_method_id(mrb, file, MRB_SYM(unlink), mrb_file_s_unlink, MRB_ARGS_ANY());
+  mrb_define_class_method_id(mrb, file, MRB_SYM(rename), mrb_file_s_rename, MRB_ARGS_REQ(2));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(symlink), mrb_file_s_symlink, MRB_ARGS_REQ(2));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(chmod), mrb_file_s_chmod, MRB_ARGS_REQ(1) | MRB_ARGS_REST());
+  mrb_define_class_method_id(mrb, file, MRB_SYM(readlink), mrb_file_s_readlink, MRB_ARGS_REQ(1));
 
-  mrb_define_class_method(mrb, file, "dirname",   mrb_file_dirname,    MRB_ARGS_REQ(1));
-  mrb_define_class_method(mrb, file, "basename",  mrb_file_basename,   MRB_ARGS_REQ(1));
-  mrb_define_class_method(mrb, file, "realpath",  mrb_file_realpath,   MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
-  mrb_define_class_method(mrb, file, "_getwd",    mrb_file__getwd,     MRB_ARGS_NONE());
-  mrb_define_class_method(mrb, file, "_gethome",  mrb_file__gethome,   MRB_ARGS_OPT(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(dirname),   mrb_file_dirname,    MRB_ARGS_REQ(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(basename),  mrb_file_basename,   MRB_ARGS_REQ(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(realpath),  mrb_file_realpath,   MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(absolute_path), mrb_file_absolute_path, MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM_Q(absolute_path), mrb_file_absolute_path_p, MRB_ARGS_REQ(1));
+  mrb_define_class_method_id(mrb, file, MRB_SYM(expand_path),  mrb_file_expand_path, MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
 
-  mrb_define_method(mrb, file, "flock", mrb_file_flock, MRB_ARGS_REQ(1));
-  mrb_define_method(mrb, file, "_mtime", mrb_file_mtime, MRB_ARGS_NONE());
-  mrb_define_method(mrb, file, "size", mrb_file_size, MRB_ARGS_NONE());
-  mrb_define_method(mrb, file, "truncate", mrb_file_truncate, MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, file, MRB_SYM(flock), mrb_file_flock, MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, file, MRB_SYM(_atime), mrb_file_atime, MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, file, MRB_SYM(_ctime), mrb_file_ctime, MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, file, MRB_SYM(_mtime), mrb_file_mtime, MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, file, MRB_SYM(size), mrb_file_size, MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, file, MRB_SYM(truncate), mrb_file_truncate, MRB_ARGS_REQ(1));
 
-  cnst = mrb_define_module_under_id(mrb, file, MRB_SYM(Constants));
+  struct RClass *cnst = mrb_define_module_under_id(mrb, file, MRB_SYM(Constants));
   mrb_define_const_id(mrb, cnst, MRB_SYM(LOCK_SH), mrb_fixnum_value(LOCK_SH));
   mrb_define_const_id(mrb, cnst, MRB_SYM(LOCK_EX), mrb_fixnum_value(LOCK_EX));
   mrb_define_const_id(mrb, cnst, MRB_SYM(LOCK_UN), mrb_fixnum_value(LOCK_UN));
   mrb_define_const_id(mrb, cnst, MRB_SYM(LOCK_NB), mrb_fixnum_value(LOCK_NB));
   mrb_define_const_id(mrb, cnst, MRB_SYM(SEPARATOR), mrb_str_new_cstr(mrb, FILE_SEPARATOR));
   mrb_define_const_id(mrb, cnst, MRB_SYM(PATH_SEPARATOR), mrb_str_new_cstr(mrb, PATH_SEPARATOR));
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32)
   mrb_define_const_id(mrb, cnst, MRB_SYM(ALT_SEPARATOR), mrb_str_new_cstr(mrb, FILE_ALT_SEPARATOR));
 #else
   mrb_define_const_id(mrb, cnst, MRB_SYM(ALT_SEPARATOR), mrb_nil_value());
