@@ -3,13 +3,14 @@
  */
 
 #include <mruby.h>
-#include "mruby/error.h"
-#include "mruby/array.h"
-#include "mruby/class.h"
-#include "mruby/numeric.h"
-#include "mruby/string.h"
-#include "mruby/variable.h"
-#include "mruby/endian.h"
+#include <mruby/error.h>
+#include <mruby/array.h>
+#include <mruby/class.h>
+#include <mruby/numeric.h>
+#include <mruby/string.h>
+#include <mruby/variable.h>
+#include <mruby/endian.h>
+#include <mruby/presym.h>
 
 #include <ctype.h>
 #include <string.h>
@@ -35,17 +36,17 @@ enum pack_dir {
   //PACK_DIR_VAX,
   PACK_DIR_BER,       /* w */
   PACK_DIR_UTF8,      /* U */
-  //PACK_DIR_BER,
   PACK_DIR_DOUBLE,    /* E */
   PACK_DIR_FLOAT,     /* f */
   PACK_DIR_STR,       /* A */
   PACK_DIR_HEX,       /* h */
+  PACK_DIR_BSTR,      /* b */
   PACK_DIR_BASE64,    /* m */
   PACK_DIR_QENC,      /* M */
   PACK_DIR_NUL,       /* x */
   PACK_DIR_BACK,      /* X */
   PACK_DIR_ABS,       /* @ */
-  PACK_DIR_INVALID
+  PACK_DIR_NONE,      /* - */
 };
 
 enum pack_type {
@@ -68,18 +69,25 @@ enum pack_type {
 
 #define PACK_BASE64_IGNORE      0xff
 #define PACK_BASE64_PADDING     0xfe
+#define IGN                     PACK_BASE64_IGNORE
+#define PAD                     PACK_BASE64_PADDING
 
-const static unsigned char base64chars[] =
-    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-const static unsigned char base64_dec_tab[] =
-  "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
-  "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
-  "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x3e\xff\xff\xff\x3f"
-  "\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\xff\xff\xff\xfe\xff\xff"
-  "\xff\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e"
-  "\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\xff\xff\xff\xff\xff"
-  "\xff\x1a\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23\x24\x25\x26\x27\x28"
-  "\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\xff\xff\xff\xff\xff";
+static const unsigned char base64chars[64] = {
+  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
+  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
+  'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
+  'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/',
+};
+static const unsigned char base64_dec_tab[128] = {
+  IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN,
+  IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN,
+  IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN, IGN,  62, IGN, IGN, IGN,  63,
+   52,  53,  54,  55,  56,  57,  58,  59,  60,  61, IGN, IGN, IGN, PAD, IGN, IGN,
+  IGN,   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,
+   15,  16,  17,  18,  19,  20,  21,  22,  23,  24,  25, IGN, IGN, IGN, IGN, IGN,
+  IGN,  26,  27,  28,  29,  30,  31,  32,  33,  34,  35,  36,  37,  38,  39,  40,
+   41,  42,  43,  44,  45,  46,  47,  48,  49,  50,  51, IGN, IGN, IGN, IGN, IGN,
+};
 
 static int
 hex2int(unsigned char ch)
@@ -123,9 +131,9 @@ static int
 unpack_char(mrb_state *mrb, const void *src, int srclen, mrb_value ary, unsigned int flags)
 {
   if (flags & PACK_FLAG_SIGNED)
-    mrb_ary_push(mrb, ary, mrb_fixnum_value(*(signed char *)src));
+    mrb_ary_push(mrb, ary, mrb_fixnum_value(*(signed char*)src));
   else
-    mrb_ary_push(mrb, ary, mrb_fixnum_value(*(unsigned char *)src));
+    mrb_ary_push(mrb, ary, mrb_fixnum_value(*(unsigned char*)src));
   return 1;
 }
 
@@ -139,7 +147,8 @@ pack_short(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned in
   if (flags & PACK_FLAG_LITTLEENDIAN) {
     RSTRING_PTR(str)[sidx+0] = n % 256;
     RSTRING_PTR(str)[sidx+1] = n / 256;
-  } else {
+  }
+  else {
     RSTRING_PTR(str)[sidx+0] = n / 256;
     RSTRING_PTR(str)[sidx+1] = n % 256;
   }
@@ -153,7 +162,8 @@ unpack_short(mrb_state *mrb, const unsigned char *src, int srclen, mrb_value ary
 
   if (flags & PACK_FLAG_LITTLEENDIAN) {
     n = src[1] * 256 + src[0];
-  } else {
+  }
+  else {
     n = src[0] * 256 + src[1];
   }
   if ((flags & PACK_FLAG_SIGNED) && (n >= 0x8000)) {
@@ -175,7 +185,8 @@ pack_long(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned int
     RSTRING_PTR(str)[sidx+1] = (char)(n >> 8);
     RSTRING_PTR(str)[sidx+2] = (char)(n >> 16);
     RSTRING_PTR(str)[sidx+3] = (char)(n >> 24);
-  } else {
+  }
+  else {
     RSTRING_PTR(str)[sidx+0] = (char)(n >> 24);
     RSTRING_PTR(str)[sidx+1] = (char)(n >> 16);
     RSTRING_PTR(str)[sidx+2] = (char)(n >> 8);
@@ -231,7 +242,8 @@ unpack_long(mrb_state *mrb, const unsigned char *src, int srclen, mrb_value ary,
     ul += (uint32_t)src[2] *256*256;
     ul += (uint32_t)src[1] *256;
     ul += (uint32_t)src[0];
-  } else {
+  }
+  else {
     ul = (uint32_t)src[0] * 256*256*256;
     ul += (uint32_t)src[1] *256*256;
     ul += (uint32_t)src[2] *256;
@@ -239,7 +251,8 @@ unpack_long(mrb_state *mrb, const unsigned char *src, int srclen, mrb_value ary,
   }
   if (flags & PACK_FLAG_SIGNED) {
     n = (int32_t)ul;
-  } else {
+  }
+  else {
 #ifndef MRB_INT64
     if (UINT_OVERFLOW_P(ul)) {
       u32tostr(msg, sizeof(msg), ul);
@@ -268,7 +281,8 @@ pack_quad(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned int
     RSTRING_PTR(str)[sidx+5] = (char)(n >> 40);
     RSTRING_PTR(str)[sidx+6] = (char)(n >> 48);
     RSTRING_PTR(str)[sidx+7] = (char)(n >> 56);
-  } else {
+  }
+  else {
     RSTRING_PTR(str)[sidx+0] = (char)(n >> 56);
     RSTRING_PTR(str)[sidx+1] = (char)(n >> 48);
     RSTRING_PTR(str)[sidx+2] = (char)(n >> 40);
@@ -285,27 +299,24 @@ static void
 u64tostr(char *buf, size_t len, uint64_t n)
 {
 #ifdef MRB_NO_STDIO
-  char *bufend = buf + len;
-  char *p = bufend - 1;
+  mrb_assert(len > 0);
 
-  if (len < 1) {
+  if (n < 10) {
+    buf[0] = '0' + n;
+    buf[1] = '\0';
     return;
   }
 
+  char *bufend = buf + len;
+  char *p = bufend - 1;
+
   *p-- = '\0';
   len--;
 
-  if (n > 0) {
-    for (; len > 0 && n > 0; len--, n /= 10) {
-      *p-- = '0' + (n % 10);
-    }
-    p++;
+  for (; len > 0 && n > 0; len--, n /= 10) {
+    *p-- = '0' + (n % 10);
   }
-  else if (len > 0) {
-    *p = '0';
-    len--;
-  }
-
+  p++;
   memmove(buf, p, bufend - p);
 #else
   snprintf(buf, len, "%" PRIu64, n);
@@ -317,9 +328,7 @@ static void
 i64tostr(char *buf, size_t len, int64_t n)
 {
 #ifdef MRB_NO_STDIO
-  if (len < 1) {
-    return;
-  }
+  mrb_assert(len > 0);
 
   if (n < 0) {
     *buf++ = '-';
@@ -345,7 +354,8 @@ unpack_quad(mrb_state *mrb, const unsigned char *src, int srclen, mrb_value ary,
   if (flags & PACK_FLAG_LITTLEENDIAN) {
     pos  = 7;
     step = -1;
-  } else {
+  }
+  else {
     pos  = 0;
     step = 1;
   }
@@ -363,7 +373,8 @@ unpack_quad(mrb_state *mrb, const unsigned char *src, int srclen, mrb_value ary,
     }
 #endif
     n = (mrb_int)sll;
-  } else {
+  }
+  else {
     if (UINT_OVERFLOW_P(ull)) {
       u64tostr(msg, sizeof(msg), ull);
       mrb_raisef(mrb, E_RANGE_ERROR, "cannot unpack to Integer: %s", msg);
@@ -425,7 +436,7 @@ pack_double(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned i
 {
   int i;
   double d;
-  uint8_t *buffer = (uint8_t *)&d;
+  uint8_t *buffer = (uint8_t*)&d;
   str = str_len_ensure(mrb, str, sidx + 8);
   d = mrb_float(o);
 
@@ -438,7 +449,8 @@ pack_double(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned i
         RSTRING_PTR(str)[sidx + i] = buffer[8 - i - 1];
       }
     }
-  } else {
+  }
+  else {
     if (littleendian) {
       for (i = 0; i < 8; i++) {
         RSTRING_PTR(str)[sidx + i] = buffer[8 - i - 1];
@@ -457,7 +469,7 @@ unpack_double(mrb_state *mrb, const unsigned char * src, int srclen, mrb_value a
 {
   int i;
   double d;
-  uint8_t *buffer = (uint8_t *)&d;
+  uint8_t *buffer = (uint8_t*)&d;
 
   if (flags & PACK_FLAG_LITTLEENDIAN) {
     if (littleendian) {
@@ -468,7 +480,8 @@ unpack_double(mrb_state *mrb, const unsigned char * src, int srclen, mrb_value a
         buffer[8 - i - 1] = src[i];
       }
     }
-  } else {
+  }
+  else {
     if (littleendian) {
       for (i = 0; i < 8; i++) {
         buffer[8 - i - 1] = src[i];
@@ -488,7 +501,7 @@ pack_float(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned in
 {
   int i;
   float f;
-  uint8_t *buffer = (uint8_t *)&f;
+  uint8_t *buffer = (uint8_t*)&f;
   str = str_len_ensure(mrb, str, sidx + 4);
   f = (float)mrb_float(o);
 
@@ -501,7 +514,8 @@ pack_float(mrb_state *mrb, mrb_value o, mrb_value str, mrb_int sidx, unsigned in
         RSTRING_PTR(str)[sidx + i] = buffer[4 - i - 1];
       }
     }
-  } else {
+  }
+  else {
     if (littleendian) {
       for (i = 0; i < 4; i++) {
         RSTRING_PTR(str)[sidx + i] = buffer[4 - i - 1];
@@ -520,7 +534,7 @@ unpack_float(mrb_state *mrb, const unsigned char * src, int srclen, mrb_value ar
 {
   int i;
   float f;
-  uint8_t *buffer = (uint8_t *)&f;
+  uint8_t *buffer = (uint8_t*)&f;
 
   if (flags & PACK_FLAG_LITTLEENDIAN) {
     if (littleendian) {
@@ -531,7 +545,8 @@ unpack_float(mrb_state *mrb, const unsigned char * src, int srclen, mrb_value ar
         buffer[4 - i - 1] = src[i];
       }
     }
-  } else {
+  }
+  else {
     if (littleendian) {
       for (i = 0; i < 4; i++) {
         buffer[4 - i - 1] = src[i];
@@ -660,7 +675,7 @@ unpack_utf8(mrb_state *mrb, const unsigned char * src, int srclen, mrb_value ary
   if (srclen == 0) {
     return 1;
   }
-  uv = utf8_to_uv(mrb, (const char *)src, &lenp);
+  uv = utf8_to_uv(mrb, (const char*)src, &lenp);
   mrb_ary_push(mrb, ary, mrb_fixnum_value((mrb_int)uv));
   return (int)lenp;
 }
@@ -681,13 +696,16 @@ pack_str(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int count,
 
   if (count == 0) {
     return 0;
-  } else if (count == -1) {
+  }
+  else if (count == -1) {
     copylen = slen;
     padlen = (flags & PACK_FLAG_Z) ? 1 : 0;
-  } else if (count < slen) {
+  }
+  else if (count < slen) {
     copylen = count;
     padlen = 0;
-  } else {
+  }
+  else {
     copylen = slen;
     padlen = count - slen;
   }
@@ -716,17 +734,19 @@ unpack_str(mrb_state *mrb, const void *src, int slen, mrb_value ary, int count,
   CHECK_UNPACK_LEN(mrb, slen, ary);
 
   mrb_value dst;
-  const char *cp, *sptr;
+  const char *sptr;
   int copylen;
 
-  sptr = (const char *)src;
+  sptr = (const char*)src;
   if (count != -1 && count < slen)  {
     slen = count;
   }
   copylen = slen;
 
   if (slen >= 0 && flags & PACK_FLAG_Z) {  /* "Z" */
-    if ((cp = (const char *)memchr(sptr, '\0', slen)) != NULL) {
+    const char *cp;
+
+    if ((cp = (const char*)memchr(sptr, '\0', slen)) != NULL) {
       copylen = (int)(cp - sptr);
       if (count == -1) {
         slen = copylen + 1;
@@ -749,7 +769,6 @@ unpack_str(mrb_state *mrb, const void *src, int slen, mrb_value ary, int count,
 static int
 pack_hex(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int count, unsigned int flags)
 {
-  int a, b;
   unsigned int ashift, bshift;
   long slen;
   char *dptr, *dptr0, *sptr;
@@ -760,14 +779,16 @@ pack_hex(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int count,
   if (flags & PACK_FLAG_LSB) {
     ashift = 0;
     bshift = 4;
-  } else {
+  }
+  else {
     ashift = 4;
     bshift = 0;
   }
 
   if (count == -1) {
     count = slen;
-  } else if (slen > count) {
+  }
+  else if (slen > count) {
     slen = count;
   }
 
@@ -776,7 +797,8 @@ pack_hex(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int count,
 
   dptr0 = dptr;
   for (; count > 0; count -= 2) {
-    a = b = 0;
+    int a = 0, b = 0;
+
     if (slen > 0) {
       a = hex2int(*sptr++);
       if (a < 0) break;
@@ -807,12 +829,13 @@ unpack_hex(mrb_state *mrb, const void *src, int slen, mrb_value ary, int count,
   if (flags & PACK_FLAG_LSB) {
     ashift = 0;
     bshift = 4;
-  } else {
+  }
+  else {
     ashift = 4;
     bshift = 0;
   }
 
-  sptr = (const char *)src;
+  sptr = (const char*)src;
 
   if (count == -1)
     count = slen * 2;
@@ -842,6 +865,92 @@ unpack_hex(mrb_state *mrb, const void *src, int slen, mrb_value ary, int count,
   return (int)(sptr - sptr0);
 }
 
+static int
+pack_bstr(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int count, unsigned int flags)
+{
+  const char *sptr = RSTRING_PTR(src);
+  int slen = (int)RSTRING_LEN(src);
+
+  if (count == -1) {
+    count = slen;
+  }
+  else if (slen > count) {
+    slen = count;
+  }
+
+  dst = str_len_ensure(mrb, dst, didx + count);
+  char *dptr = RSTRING_PTR(dst) + didx;
+  char *dptr0 = dptr;
+
+  unsigned int byte = 0;
+  for (int i=0; i++ < slen; sptr++) {
+    if (flags & PACK_FLAG_LSB) {
+      if (*sptr & 1)
+        byte |= 128;
+      if (i & 7)
+        byte >>= 1;
+      else {
+        char c = (char)(byte&0xff);
+        *dptr++ = c;
+        byte = 0;
+      }
+    }
+    else {
+      byte |= *sptr & 1;
+      if (i & 7)
+        byte <<= 1;
+      else {
+        char c = (char)(byte&0xff);
+        *dptr++ = c;
+        byte = 0;
+      }
+    }
+  }
+  if (slen & 7) {
+    if (flags & PACK_FLAG_LSB) {
+      byte >>= 7 - (slen & 7);
+    }
+    else {
+      byte <<= 7 - (slen & 7);
+    }
+    char c = (char)(byte&0xff);
+    *dptr++ = c;
+  }
+  return (int)(dptr - dptr0);
+}
+
+static int
+unpack_bstr(mrb_state *mrb, const void *src, int slen, mrb_value ary, int count, unsigned int flags)
+{
+  CHECK_UNPACK_LEN(mrb, slen, ary);
+
+  const char *sptr0 = (const char*)src;
+  const char *sptr = sptr0;
+  if (count == -1 || count > slen * 8)
+    count = slen * 8;
+
+  mrb_value dst = mrb_str_new(mrb, NULL, count);
+  char *dptr = RSTRING_PTR(dst);
+  const char *dptr0 = dptr;
+  int bits = 0;
+
+  for (int i=0; i<count; i++) {
+    if (flags & PACK_FLAG_LSB) {
+      if (i & 7) bits >>= 1;
+      else bits = (unsigned char)*sptr++;
+      *dptr++ = (bits & 1) ? '1' : '0';
+    }
+    else {
+      if (i & 7) bits <<= 1;
+      else bits = (unsigned char)*sptr++;
+      *dptr++ = (bits & 128) ? '1' : '0';
+    }
+  }
+  dst = mrb_str_resize(mrb, dst, (mrb_int)(dptr - dptr0));
+  mrb_ary_push(mrb, ary, dst);
+  return (int)(sptr - sptr0);
+}
+
 static int
 pack_base64(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int count)
 {
@@ -858,7 +967,8 @@ pack_base64(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int coun
 
   if (count != 0 && count < 3) {  /* -1, 1 or 2 */
     count = 45;
-  } else if (count >= 3) {
+  }
+  else if (count >= 3) {
     count -= count % 3;
   }
 
@@ -892,7 +1002,8 @@ pack_base64(mrb_state *mrb, mrb_value src, mrb_value dst, mrb_int didx, int coun
     *dstptr++ = '=';
     *dstptr++ = '=';
     column += 3;
-  } else if (srclen == 2) {
+  }
+  else if (srclen == 2) {
     l = (unsigned char)*srcptr++ << 16;
     l += (unsigned char)*srcptr++ << 8;
     *dstptr++ = base64chars[(l >> 18) & 0x3f];
@@ -921,7 +1032,7 @@ unpack_base64(mrb_state *mrb, const void *src, int slen, mrb_value ary)
   const char *sptr, *sptr0;
   char *dptr, *dptr0;
 
-  sptr0 = sptr = (const char *)src;
+  sptr0 = sptr = (const char*)src;
 
   dlen = slen / 4 * 3;  /* an estimated value - may be shorter */
   dst = mrb_str_new(mrb, NULL, dlen);
@@ -950,11 +1061,13 @@ unpack_base64(mrb_state *mrb, const void *src, int slen, mrb_value ary)
       *dptr++ = (l >> 16) & 0xff;
       *dptr++ = (l >> 8) & 0xff;
       *dptr++ = l & 0xff;
-    } else if (padding == 1) {
+    }
+    else if (padding == 1) {
       *dptr++ = (l >> 16) & 0xff;
       *dptr++ = (l >> 8) & 0xff;
       break;
-    } else {
+    }
+    else {
       *dptr++ = (l >> 16) & 0xff;
       break;
     }
@@ -1095,8 +1208,8 @@ has_tmpl(const struct tmpl *tmpl)
   return (tmpl->idx < RSTRING_LEN(tmpl->str));
 }
 
-static void
-read_tmpl(mrb_state *mrb, struct tmpl *tmpl, enum pack_dir *dirp, enum pack_type *typep, int *sizep, int *countp, unsigned int *flagsp)
+static enum pack_dir
+read_tmpl(mrb_state *mrb, struct tmpl *tmpl, enum pack_type *typep, int *sizep, int *countp, unsigned int *flagsp)
 {
   mrb_int t, tlen;
   int ch, size = 0;
@@ -1109,8 +1222,10 @@ read_tmpl(mrb_state *mrb, struct tmpl *tmpl, enum pack_dir *dirp, enum pack_type
   tptr = RSTRING_PTR(tmpl->str);
   tlen = RSTRING_LEN(tmpl->str);
 
+ restart:
+  if (tmpl->idx >= tlen) return PACK_DIR_NONE;
   t = tptr[tmpl->idx++];
-alias:
+ alias:
   switch (t) {
   case 'A':
     dir = PACK_DIR_STR;
@@ -1179,6 +1294,16 @@ alias:
     type = PACK_TYPE_STRING;
     flags |= PACK_FLAG_COUNT2 | PACK_FLAG_LSB;
     break;
+  case 'B':
+    dir = PACK_DIR_BSTR;
+    type = PACK_TYPE_STRING;
+    flags |= PACK_FLAG_COUNT2;
+    break;
+  case 'b':
+    dir = PACK_DIR_BSTR;
+    type = PACK_TYPE_STRING;
+    flags |= PACK_FLAG_COUNT2 | PACK_FLAG_LSB;
+    break;
   case 'I':
     switch (sizeof(int)) {
       case 2: t = 'S'; goto alias;
@@ -1197,6 +1322,22 @@ alias:
         mrb_raisef(mrb, E_RUNTIME_ERROR, "mruby-pack does not support sizeof(int) == %d", (int)sizeof(int));
     }
     break;
+  case 'J':
+    switch (sizeof(intptr_t)) {
+      case 4: t = 'L'; goto alias;
+      case 8: t = 'Q'; goto alias;
+      default:
+        mrb_raisef(mrb, E_RUNTIME_ERROR, "mruby-pack does not support sizeof(uintptr_t) == %d", (int)sizeof(uintptr_t));
+    }
+    break;
+  case 'j':
+    switch (sizeof(intptr_t)) {
+      case 4: t = 'l'; goto alias;
+      case 8: t = 'q'; goto alias;
+      default:
+        mrb_raisef(mrb, E_RUNTIME_ERROR, "mruby-pack does not support sizeof(intptr_t) == %d", (int)sizeof(intptr_t));
+    }
+    break;
   case 'L':
     dir = PACK_DIR_LONG;
     type = PACK_TYPE_INTEGER;
@@ -1290,14 +1431,22 @@ alias:
     type = PACK_TYPE_STRING;
     flags |= PACK_FLAG_WIDTH | PACK_FLAG_COUNT2 | PACK_FLAG_Z;
     break;
+  case '#':
+    while (++tmpl->idx < tlen && tptr[tmpl->idx] != '\n')
+      ;
+    goto restart;
+
   case 'p': case 'P':
   case '%':
     mrb_raisef(mrb, E_ARGUMENT_ERROR, "%c is not supported", (char)t);
     break;
   default:
-    dir = PACK_DIR_INVALID;
-    type = PACK_TYPE_NONE;
-    break;
+    if (!ISSPACE((char)t)) {
+      char c = (char)t;
+      mrb_value s = mrb_str_new(mrb, &c, 1);
+      mrb_raisef(mrb, E_ARGUMENT_ERROR, "unknown unpack directive %!v", s);
+    }
+    goto restart;
   }
 
   /* read suffix [0-9*_!<>] */
@@ -1312,20 +1461,24 @@ alias:
       count = (int)n;
       tmpl->idx = (int)(e - tptr);
       continue;
-    } else if (ch == '*')  {
+    }
+    else if (ch == '*')  {
       if (type == PACK_TYPE_NONE)
         count = 0;
       else
         count = -1;
-    } else if (ch == '_' || ch == '!' || ch == '<' || ch == '>') {
+    }
+    else if (ch == '_' || ch == '!' || ch == '<' || ch == '>') {
       if (strchr("sSiIlLqQ", (int)t) == NULL) {
         mrb_raisef(mrb, E_ARGUMENT_ERROR, "'%c' allowed only after types sSiIlLqQ", ch);
       }
       if (ch == '_' || ch == '!') {
         flags |= PACK_FLAG_s;
-      } else if (ch == '<') {
+      }
+      else if (ch == '<') {
         flags |= PACK_FLAG_LT;
-      } else if (ch == '>') {
+      }
+      else if (ch == '>') {
         flags |= PACK_FLAG_GT;
       }
     }
@@ -1339,36 +1492,33 @@ alias:
     flags |= PACK_FLAG_LITTLEENDIAN;
   }
 
-  *dirp = dir;
   *typep = type;
   *sizep = size;
   *countp = count;
   *flagsp = flags;
+  return dir;
 }
 
 static mrb_value
 mrb_pack_pack(mrb_state *mrb, mrb_value ary)
 {
   mrb_value o, result;
-  mrb_int aidx;
   struct tmpl tmpl;
-  int count;
+  enum pack_type type;
+  int count, size;
   unsigned int flags;
   enum pack_dir dir;
-  enum pack_type type;
-  int ridx, size;
 
   prepare_tmpl(mrb, &tmpl);
 
   result = mrb_str_new(mrb, NULL, 128);  /* allocate initial buffer */
-  aidx = 0;
-  ridx = 0;
+  mrb_int aidx = 0;
+  mrb_int ridx = 0;
   while (has_tmpl(&tmpl)) {
-    read_tmpl(mrb, &tmpl, &dir, &type, &size, &count, &flags);
+    dir = read_tmpl(mrb, &tmpl, &type, &size, &count, &flags);
 
-    if (dir == PACK_DIR_INVALID)
-      continue;
-    else if (dir == PACK_DIR_NUL) {
+    if (dir == PACK_DIR_NONE) break;
+    if (dir == PACK_DIR_NUL) {
     grow:
       if (ridx > INT_MAX - count) goto overflow;
       ridx += pack_nul(mrb, result, ridx, count);
@@ -1395,7 +1545,7 @@ mrb_pack_pack(mrb_state *mrb, mrb_value ary)
       if (count == 0 && !(flags & PACK_FLAG_WIDTH))
         break;
 
-      o = mrb_ary_ref(mrb, ary, aidx);
+      o = RARRAY_PTR(ary)[aidx];
       if (type == PACK_TYPE_INTEGER) {
         o = mrb_ensure_int_type(mrb, o);
       }
@@ -1437,6 +1587,9 @@ mrb_pack_pack(mrb_state *mrb, mrb_value ary)
       case PACK_DIR_HEX:
         ridx += pack_hex(mrb, o, result, ridx, count, flags);
         break;
+      case PACK_DIR_BSTR:
+        ridx += pack_bstr(mrb, o, result, ridx, count, flags);
+        break;
       case PACK_DIR_STR:
         ridx += pack_str(mrb, o, result, ridx, count, flags);
         break;
@@ -1493,11 +1646,10 @@ pack_unpack(mrb_state *mrb, mrb_value str, int single)
 
   result = mrb_ary_new(mrb);
   while (has_tmpl(&tmpl)) {
-    read_tmpl(mrb, &tmpl, &dir, &type, &size, &count, &flags);
+    dir = read_tmpl(mrb, &tmpl, &type, &size, &count, &flags);
 
-    if (dir == PACK_DIR_INVALID)
-      continue;
-    else if (dir == PACK_DIR_NUL) {
+    if (dir == PACK_DIR_NONE) break;
+    if (dir == PACK_DIR_NUL) {
       check_x(mrb, srclen-srcidx, count, 'x');
       srcidx += count;
       continue;
@@ -1514,19 +1666,27 @@ pack_unpack(mrb_state *mrb, mrb_value str, int single)
     }
 
     /* PACK_FLAG_COUNT2 directions */
-    sptr = (const unsigned char *)RSTRING_PTR(str) + srcidx;
+    sptr = (const unsigned char*)RSTRING_PTR(str) + srcidx;
     switch (dir) {
     case PACK_DIR_HEX:
       srcidx += unpack_hex(mrb, sptr, srclen - srcidx, result, count, flags);
+      if (single) goto single_return;
+      continue;
+    case PACK_DIR_BSTR:
+      srcidx += unpack_bstr(mrb, sptr, srclen - srcidx, result, count, flags);
+      if (single) goto single_return;
       continue;
     case PACK_DIR_STR:
       srcidx += unpack_str(mrb, sptr, srclen - srcidx, result, count, flags);
+      if (single) goto single_return;
       continue;
     case PACK_DIR_BASE64:
       srcidx += unpack_base64(mrb, sptr, srclen - srcidx, result);
+      if (single) goto single_return;
       continue;
     case PACK_DIR_QENC:
       srcidx += unpack_qenc(mrb, sptr, srclen - srcidx, result);
+      if (single) goto single_return;
       continue;
     default:
       break;
@@ -1575,12 +1735,13 @@ pack_unpack(mrb_state *mrb, mrb_value str, int single)
         count--;
       }
     }
-    if (single) {
-      if (RARRAY_LEN(result) > 0) {
-        return RARRAY_PTR(result)[0];
-      }
-      return mrb_nil_value();
+  }
+  if (single) {
+  single_return:
+    if (RARRAY_LEN(result) > 0) {
+      return RARRAY_PTR(result)[0];
     }
+    return mrb_nil_value();
   }
   return result;
 }
@@ -1600,9 +1761,9 @@ mrb_pack_unpack1(mrb_state *mrb, mrb_value str)
 void
 mrb_mruby_pack_gem_init(mrb_state *mrb)
 {
-  mrb_define_method(mrb, mrb->array_class, "pack", mrb_pack_pack, MRB_ARGS_REQ(1));
-  mrb_define_method(mrb, mrb->string_class, "unpack", mrb_pack_unpack, MRB_ARGS_REQ(1));
-  mrb_define_method(mrb, mrb->string_class, "unpack1", mrb_pack_unpack1, MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, mrb->array_class, MRB_SYM(pack), mrb_pack_pack, MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, mrb->string_class, MRB_SYM(unpack), mrb_pack_unpack, MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, mrb->string_class, MRB_SYM(unpack1), mrb_pack_unpack1, MRB_ARGS_REQ(1));
 }
 
 void
