@@ -13,62 +13,60 @@
 #include <mruby/hash.h>
 #include <mruby/range.h>
 #include <mruby/proc.h>
+#include <mruby/internal.h>
+#include <mruby/presym.h>
 
 #define RSTRUCT_LEN(st) RARRAY_LEN(st)
 #define RSTRUCT_PTR(st) RARRAY_PTR(st)
 
+#define mrb_struct_p(o) (mrb_type(o) == MRB_TT_STRUCT)
+
 static struct RClass *
 struct_class(mrb_state *mrb)
 {
-  return mrb_class_get(mrb, "Struct");
+  return mrb_class_get_id(mrb, MRB_SYM(Struct));
 }
 
-static inline mrb_value
-struct_ivar_get(mrb_state *mrb, mrb_value cls, mrb_sym id)
+static void
+struct_corrupted(mrb_state *mrb)
 {
-  struct RClass* c = mrb_class_ptr(cls);
-  struct RClass* sclass = struct_class(mrb);
-  mrb_value ans;
-
-  for (;;) {
-    ans = mrb_iv_get(mrb, mrb_obj_value(c), id);
-    if (!mrb_nil_p(ans)) return ans;
-    c = c->super;
-    if (c == sclass || c == 0)
-      return mrb_nil_value();
-  }
+  mrb_raise(mrb, E_TYPE_ERROR, "corrupted struct");
 }
 
 static mrb_value
-struct_s_members(mrb_state *mrb, struct RClass *klass)
+struct_s_members(mrb_state *mrb, struct RClass *c)
 {
-  mrb_value members = struct_ivar_get(mrb, mrb_obj_value(klass), mrb_intern_lit(mrb, "__members__"));
+  struct RClass* sclass = struct_class(mrb);
 
-  if (mrb_nil_p(members)) {
-    mrb_raise(mrb, E_TYPE_ERROR, "uninitialized struct");
+  for (;;) {
+    mrb_value members = mrb_iv_get(mrb, mrb_obj_value(c), MRB_SYM(__members__));
+
+    if (!mrb_nil_p(members)) {
+      if (!mrb_array_p(members)) {
+        struct_corrupted(mrb);
+      }
+      return members;
+    }
+    c = c->super;
+    if (c == sclass || c == 0) {
+      mrb_raise(mrb, E_TYPE_ERROR, "uninitialized struct");
+    }
   }
-  if (!mrb_array_p(members)) {
-    mrb_raise(mrb, E_TYPE_ERROR, "corrupted struct");
-  }
-  return members;
 }
 
 static mrb_value
 struct_members(mrb_state *mrb, mrb_value s)
 {
-  mrb_value members = struct_s_members(mrb, mrb_obj_class(mrb, s));
-  if (!mrb_array_p(s)) {
-    mrb_raise(mrb, E_TYPE_ERROR, "corrupted struct");
+  if (!mrb_struct_p(s)) {
+    struct_corrupted(mrb);
   }
-  if (RSTRUCT_LEN(s) != RARRAY_LEN(members)) {
-    if (RSTRUCT_LEN(s) == 0) {  /* probably uninitialized */
-      mrb_ary_resize(mrb, s, RARRAY_LEN(members));
-    }
-    else {
-      mrb_raisef(mrb, E_TYPE_ERROR,
-                 "struct size differs (%S required %S given)",
-                 mrb_fixnum_value(RARRAY_LEN(members)), mrb_fixnum_value(RSTRUCT_LEN(s)));
-    }
+  mrb_value members = struct_s_members(mrb, mrb_obj_class(mrb, s));
+  mrb_int len = RSTRUCT_LEN(s);
+  mrb_int mlen = RARRAY_LEN(members);
+  if (len > 0 && len != mlen) {
+    mrb_raisef(mrb, E_TYPE_ERROR,
+               "struct size differs (%i required %i given)",
+               mlen, len);
   }
   return members;
 }
@@ -76,23 +74,13 @@ struct_members(mrb_state *mrb, mrb_value s)
 static mrb_value
 mrb_struct_s_members_m(mrb_state *mrb, mrb_value klass)
 {
-  mrb_value members, ary;
-
-  members = struct_s_members(mrb, mrb_class_ptr(klass));
-  ary = mrb_ary_new_capa(mrb, RARRAY_LEN(members));
+  mrb_value members = struct_s_members(mrb, mrb_class_ptr(klass));
+  mrb_value ary = mrb_ary_new_capa(mrb, RARRAY_LEN(members));
   mrb_ary_replace(mrb, ary, members);
   return ary;
 }
 
-static void
-mrb_struct_modify(mrb_state *mrb, mrb_value strct)
-{
-  if (MRB_FROZEN_P(mrb_basic_ptr(strct))) {
-    mrb_raise(mrb, E_FROZEN_ERROR, "can't modify frozen struct");
-  }
-
-  mrb_write_barrier(mrb, mrb_basic_ptr(strct));
-}
+#define mrb_struct_modify(mrb,s) mrb_check_frozen((mrb), mrb_basic_ptr(s))
 
 /* 15.2.18.4.6  */
 /*
@@ -113,99 +101,89 @@ mrb_struct_members(mrb_state *mrb, mrb_value obj)
   return mrb_struct_s_members_m(mrb, mrb_obj_value(mrb_obj_class(mrb, obj)));
 }
 
+static mrb_int
+num_members(mrb_state *mrb, mrb_value self)
+{
+  mrb_value members = struct_members(mrb, self);
+  return RARRAY_LEN(members);
+}
+
 static mrb_value
 mrb_struct_ref(mrb_state *mrb, mrb_value obj)
 {
-  mrb_int i = mrb_fixnum(mrb_proc_cfunc_env_get(mrb, 0));
+  mrb_int argc = mrb_get_argc(mrb);
+  if (argc != 0) {
+    mrb_argnum_error(mrb, argc, 0, 0);
+  }
+  mrb_int i = mrb_integer(mrb_proc_cfunc_env_get(mrb, 0));
+  mrb_int len = num_members(mrb, obj);
   mrb_value *ptr = RSTRUCT_PTR(obj);
 
-  if (!ptr) return mrb_nil_value();
+  if (!ptr || len <= i) return mrb_nil_value();
   return ptr[i];
 }
 
 static mrb_sym
 mrb_id_attrset(mrb_state *mrb, mrb_sym id)
 {
-  const char *name;
+#define ONSTACK_ALLOC_MAX 32
+#define ONSTACK_STRLEN_MAX (ONSTACK_ALLOC_MAX - 1) /* '=' character */
+
   char *buf;
   mrb_int len;
-  mrb_sym mid;
+  char onstack[ONSTACK_ALLOC_MAX];
 
-  name = mrb_sym2name_len(mrb, id, &len);
-  buf = (char *)mrb_malloc(mrb, (size_t)len+2);
+  const char *name = mrb_sym_name_len(mrb, id, &len);
+  if (len > ONSTACK_STRLEN_MAX) {
+    buf = (char*)mrb_malloc(mrb, (size_t)len+1);
+  }
+  else {
+    buf = onstack;
+  }
   memcpy(buf, name, (size_t)len);
   buf[len] = '=';
-  buf[len+1] = '\0';
 
-  mid = mrb_intern(mrb, buf, len+1);
-  mrb_free(mrb, buf);
+  mrb_sym mid = mrb_intern(mrb, buf, len+1);
+  if (buf != onstack) {
+    mrb_free(mrb, buf);
+  }
   return mid;
 }
 
 static mrb_value
 mrb_struct_set_m(mrb_state *mrb, mrb_value obj)
 {
-  mrb_int i = mrb_fixnum(mrb_proc_cfunc_env_get(mrb, 0));
-  mrb_value *ptr;
-  mrb_value val;
+  mrb_int i = mrb_integer(mrb_proc_cfunc_env_get(mrb, 0));
+  mrb_value val = mrb_get_arg1(mrb);
 
-  mrb_get_args(mrb, "o", &val);
-  mrb_struct_modify(mrb, obj);
-  ptr = RSTRUCT_PTR(obj);
-  if (ptr == NULL || i >= RSTRUCT_LEN(obj)) {
-    mrb_ary_set(mrb, obj, i, val);
-  }
-  else {
-    ptr[i] = val;
-  }
+  mrb_ary_set(mrb, obj, i, val);
   return val;
 }
 
-static mrb_bool
-is_local_id(mrb_state *mrb, const char *name)
-{
-  if (!name) return FALSE;
-  return !ISUPPER(name[0]);
-}
-
-static mrb_bool
-is_const_id(mrb_state *mrb, const char *name)
-{
-  if (!name) return FALSE;
-  return ISUPPER(name[0]);
-}
-
 static void
 make_struct_define_accessors(mrb_state *mrb, mrb_value members, struct RClass *c)
 {
   const mrb_value *ptr_members = RARRAY_PTR(members);
-  mrb_int i;
   mrb_int len = RARRAY_LEN(members);
   int ai = mrb_gc_arena_save(mrb);
 
-  for (i=0; i<len; i++) {
+  for (mrb_int i=0; i<len; i++) {
     mrb_sym id = mrb_symbol(ptr_members[i]);
-    const char *name = mrb_sym2name_len(mrb, id, NULL);
-
-    if (is_local_id(mrb, name) || is_const_id(mrb, name)) {
-      mrb_method_t m;
-      mrb_value at = mrb_fixnum_value(i);
-      struct RProc *aref = mrb_proc_new_cfunc_with_env(mrb, mrb_struct_ref, 1, &at);
-      struct RProc *aset = mrb_proc_new_cfunc_with_env(mrb, mrb_struct_set_m, 1, &at);
-      MRB_METHOD_FROM_PROC(m, aref);
-      mrb_define_method_raw(mrb, c, id, m);
-      MRB_METHOD_FROM_PROC(m, aset);
-      mrb_define_method_raw(mrb, c, mrb_id_attrset(mrb, id), m);
-      mrb_gc_arena_restore(mrb, ai);
-    }
+    mrb_method_t m;
+    mrb_value at = mrb_fixnum_value(i);
+    struct RProc *aref = mrb_proc_new_cfunc_with_env(mrb, mrb_struct_ref, 1, &at);
+    struct RProc *aset = mrb_proc_new_cfunc_with_env(mrb, mrb_struct_set_m, 1, &at);
+    MRB_METHOD_FROM_PROC(m, aref);
+    mrb_define_method_raw(mrb, c, id, m);
+    MRB_METHOD_FROM_PROC(m, aset);
+    mrb_define_method_raw(mrb, c, mrb_id_attrset(mrb, id), m);
+    mrb_gc_arena_restore(mrb, ai);
   }
 }
 
 static mrb_value
 make_struct(mrb_state *mrb, mrb_value name, mrb_value members, struct RClass *klass)
 {
-  mrb_value nstr;
-  mrb_sym id;
   struct RClass *c;
 
   if (mrb_nil_p(name)) {
@@ -213,24 +191,27 @@ make_struct(mrb_state *mrb, mrb_value name, mrb_value members, struct RClass *kl
   }
   else {
     /* old style: should we warn? */
-    mrb_to_str(mrb, name);
-    id = mrb_obj_to_sym(mrb, name);
-    if (!is_const_id(mrb, mrb_sym2name_len(mrb, id, NULL))) {
-      mrb_name_error(mrb, id, "identifier %S needs to be constant", name);
+    mrb_ensure_string_type(mrb, name);
+
+    mrb_sym id = mrb_obj_to_sym(mrb, name);
+    if (!mrb_const_name_p(mrb, RSTRING_PTR(name), RSTRING_LEN(name))) {
+      mrb_name_error(mrb, id, "identifier %v needs to be constant", name);
     }
     if (mrb_const_defined_at(mrb, mrb_obj_value(klass), id)) {
-      mrb_warn(mrb, "redefining constant Struct::%S", name);
+      mrb_warn(mrb, "redefining constant Struct::%v", name);
       mrb_const_remove(mrb, mrb_obj_value(klass), id);
     }
-    c = mrb_define_class_under(mrb, klass, RSTRING_PTR(name), klass);
+    c = mrb_define_class_under_id(mrb, klass, id, klass);
   }
-  MRB_SET_INSTANCE_TT(c, MRB_TT_ARRAY);
-  nstr = mrb_obj_value(c);
-  mrb_iv_set(mrb, nstr, mrb_intern_lit(mrb, "__members__"), members);
+  MRB_SET_INSTANCE_TT(c, MRB_TT_STRUCT);
+  MRB_DEFINE_ALLOCATOR(c);
 
-  mrb_define_class_method(mrb, c, "new", mrb_instance_new, MRB_ARGS_ANY());
-  mrb_define_class_method(mrb, c, "[]", mrb_instance_new, MRB_ARGS_ANY());
-  mrb_define_class_method(mrb, c, "members", mrb_struct_s_members_m, MRB_ARGS_NONE());
+  mrb_value nstr = mrb_obj_value(c);
+  mrb_iv_set(mrb, nstr, MRB_SYM(__members__), members);
+
+  mrb_define_class_method_id(mrb, c, MRB_SYM(new), mrb_instance_new, MRB_ARGS_ANY());
+  mrb_define_class_method_id(mrb, c, MRB_OPSYM(aref), mrb_instance_new, MRB_ARGS_ANY());
+  mrb_define_class_method_id(mrb, c, MRB_SYM(members), mrb_struct_s_members_m, MRB_ARGS_NONE());
   /* RSTRUCT(nstr)->basic.c->super = c->c; */
   make_struct_define_accessors(mrb, members, c);
   return nstr;
@@ -273,80 +254,64 @@ make_struct(mrb_state *mrb, mrb_value name, mrb_value members, struct RClass *kl
 static mrb_value
 mrb_struct_s_def(mrb_state *mrb, mrb_value klass)
 {
-  mrb_value name, rest;
-  mrb_value *pargv;
-  mrb_int argcnt;
-  mrb_int i;
-  mrb_value b, st;
-  mrb_sym id;
-  mrb_value *argv;
+  mrb_value name = mrb_nil_value();
+  mrb_value b;
+  const mrb_value *argv;
   mrb_int argc;
 
-  name = mrb_nil_value();
   mrb_get_args(mrb, "*&", &argv, &argc, &b);
-  if (argc == 0) { /* special case to avoid crash */
-    mrb_raise(mrb, E_ARGUMENT_ERROR, "wrong number of arguments");
+  if (argc == 0) {
+    mrb_raise(mrb, E_ARGUMENT_ERROR, "wrong number of arguments (given 0, expected 1+)");
   }
-  else {
-    pargv = argv;
-    argcnt = argc;
-    if (argc > 0) {
-      name = argv[0];
-      if (mrb_symbol_p(name)) {
-        /* 1stArgument:symbol -> name=nil rest=argv[0..n] */
-        name = mrb_nil_value();
-      }
-      else {
-        pargv++;
-        argcnt--;
+
+  const mrb_value *pargv = argv;
+  mrb_int argcnt = argc;
+  if (argc > 0 && !mrb_symbol_p(argv[0])) {
+    /* 1stArgument:!symbol -> name=argv[0] rest=argv[0..n] */
+    name = argv[0];
+    pargv++;
+    argcnt--;
+  }
+  mrb_value members = mrb_ary_new_from_values(mrb, argcnt, pargv);
+  for (mrb_int i=0; i<argcnt; i++) {
+    mrb_sym sym = mrb_obj_to_sym(mrb, RARRAY_PTR(members)[i]);
+    mrb_ary_set(mrb, members, i, mrb_symbol_value(sym));
+  }
+  /* check member duplication */
+  mrb_int len = RARRAY_LEN(members);
+  mrb_value *p = RARRAY_PTR(members);
+  for (mrb_int i=0; i<len; i++) {
+    mrb_sym sym = mrb_symbol(p[i]);
+    for (mrb_int j=i+1; j<len; j++) {
+      if (sym == mrb_symbol(p[j])) {
+        mrb_raisef(mrb, E_ARGUMENT_ERROR, "duplicate member: %n", sym);
       }
     }
-    rest = mrb_ary_new_from_values(mrb, argcnt, pargv);
-    for (i=0; i<argcnt; i++) {
-      id = mrb_obj_to_sym(mrb, RARRAY_PTR(rest)[i]);
-      mrb_ary_set(mrb, rest, i, mrb_symbol_value(id));
-    }
-    st = make_struct(mrb, name, rest, mrb_class_ptr(klass));
-    if (!mrb_nil_p(b)) {
-      mrb_yield_with_class(mrb, b, 1, &st, st, mrb_class_ptr(st));
-    }
-
-    return st;
   }
-  /* not reached */
-  return mrb_nil_value();
-}
 
-static mrb_int
-num_members(mrb_state *mrb, struct RClass *klass)
-{
-  mrb_value members;
-
-  members = struct_ivar_get(mrb, mrb_obj_value(klass), mrb_intern_lit(mrb, "__members__"));
-  if (!mrb_array_p(members)) {
-    mrb_raise(mrb, E_TYPE_ERROR, "broken members");
+  mrb_value st = make_struct(mrb, name, members, mrb_class_ptr(klass));
+  if (!mrb_nil_p(b)) {
+    mrb_yield_with_class(mrb, b, 1, &st, st, mrb_class_ptr(st));
   }
-  return RARRAY_LEN(members);
+  return st;
 }
 
 /* 15.2.18.4.8  */
 /*
  */
 static mrb_value
-mrb_struct_initialize_withArg(mrb_state *mrb, mrb_int argc, mrb_value *argv, mrb_value self)
+mrb_struct_initialize_withArg(mrb_state *mrb, mrb_int argc, const mrb_value *argv, mrb_value self)
 {
-  struct RClass *klass = mrb_obj_class(mrb, self);
-  mrb_int i, n;
+  mrb_int n = num_members(mrb, self);
 
-  n = num_members(mrb, klass);
   if (n < argc) {
     mrb_raise(mrb, E_ARGUMENT_ERROR, "struct size differs");
   }
 
-  for (i = 0; i < argc; i++) {
+  for (mrb_int i = 0; i < argc; i++) {
     mrb_ary_set(mrb, self, i, argv[i]);
   }
-  for (i = argc; i < n; i++) {
+  for (mrb_int i = argc; i < n; i++) {
     mrb_ary_set(mrb, self, i, mrb_nil_value());
   }
   return self;
@@ -355,10 +320,10 @@ mrb_struct_initialize_withArg(mrb_state *mrb, mrb_int argc, mrb_value *argv, mrb
 static mrb_value
 mrb_struct_initialize(mrb_state *mrb, mrb_value self)
 {
-  mrb_value *argv;
+  const mrb_value *argv;
   mrb_int argc;
 
-  mrb_get_args(mrb, "*!", &argv, &argc);
+  mrb_get_args(mrb, "*", &argv, &argc);
   return mrb_struct_initialize_withArg(mrb, argc, argv, self);
 }
 
@@ -367,16 +332,14 @@ mrb_struct_initialize(mrb_state *mrb, mrb_value self)
 static mrb_value
 mrb_struct_init_copy(mrb_state *mrb, mrb_value copy)
 {
-  mrb_value s;
-
-  mrb_get_args(mrb, "o", &s);
+  mrb_value s = mrb_get_arg1(mrb);
 
   if (mrb_obj_equal(mrb, copy, s)) return copy;
   if (!mrb_obj_is_instance_of(mrb, s, mrb_obj_class(mrb, copy))) {
     mrb_raise(mrb, E_TYPE_ERROR, "wrong argument class");
   }
-  if (!mrb_array_p(s)) {
-    mrb_raise(mrb, E_TYPE_ERROR, "corrupted struct");
+  if (!mrb_struct_p(s)) {
+    struct_corrupted(mrb);
   }
   mrb_ary_replace(mrb, copy, s);
   return copy;
@@ -385,37 +348,42 @@ mrb_struct_init_copy(mrb_state *mrb, mrb_value copy)
 static mrb_value
 struct_aref_sym(mrb_state *mrb, mrb_value obj, mrb_sym id)
 {
-  mrb_value members, *ptr;
-  const mrb_value *ptr_members;
-  mrb_int i, len;
-
-  members = struct_members(mrb, obj);
-  ptr_members = RARRAY_PTR(members);
-  len = RARRAY_LEN(members);
-  ptr = RSTRUCT_PTR(obj);
-  for (i=0; i<len; i++) {
+  mrb_value members = struct_members(mrb, obj);
+  const mrb_value *ptr_members = RARRAY_PTR(members);
+  mrb_int len = RARRAY_LEN(members);
+  mrb_value *ptr = RSTRUCT_PTR(obj);
+  mrb_int plen = RARRAY_LEN(obj);
+  for (mrb_int i=0; i<len; i++) {
     mrb_value slot = ptr_members[i];
     if (mrb_symbol_p(slot) && mrb_symbol(slot) == id) {
-      return ptr[i];
+      if (i < plen) return ptr[i];
+      return mrb_nil_value();
     }
   }
-  mrb_raisef(mrb, E_INDEX_ERROR, "'%S' is not a struct member", mrb_sym2str(mrb, id));
+  mrb_name_error(mrb, id, "no member '%n' in struct", id);
   return mrb_nil_value();       /* not reached */
 }
 
+static mrb_int
+struct_index(mrb_state *mrb, mrb_int i, mrb_int len)
+{
+  mrb_int idx = i < 0 ? len + i : i;
+
+  if (idx < 0)
+    mrb_raisef(mrb, E_INDEX_ERROR,
+               "offset %i too small for struct(size:%i)", i, len);
+  if (len <= idx)
+    mrb_raisef(mrb, E_INDEX_ERROR,
+               "offset %i too large for struct(size:%i)", i, len);
+  return idx;
+}
+
 static mrb_value
 struct_aref_int(mrb_state *mrb, mrb_value s, mrb_int i)
 {
-  if (i < 0) i = RSTRUCT_LEN(s) + i;
-  if (i < 0)
-      mrb_raisef(mrb, E_INDEX_ERROR,
-                 "offset %S too small for struct(size:%S)",
-                 mrb_fixnum_value(i), mrb_fixnum_value(RSTRUCT_LEN(s)));
-  if (RSTRUCT_LEN(s) <= i)
-    mrb_raisef(mrb, E_INDEX_ERROR,
-               "offset %S too large for struct(size:%S)",
-               mrb_fixnum_value(i), mrb_fixnum_value(RSTRUCT_LEN(s)));
-  return RSTRUCT_PTR(s)[i];
+  mrb_int idx = struct_index(mrb, i, num_members(mrb, s));
+  if (idx >= RSTRUCT_LEN(s)) return mrb_nil_value();
+  return RSTRUCT_PTR(s)[idx];
 }
 
 /* 15.2.18.4.2  */
@@ -440,42 +408,31 @@ struct_aref_int(mrb_state *mrb, mrb_value s, mrb_int i)
 static mrb_value
 mrb_struct_aref(mrb_state *mrb, mrb_value s)
 {
-  mrb_value idx;
+  mrb_value idx = mrb_get_arg1(mrb);
 
-  mrb_get_args(mrb, "o", &idx);
   if (mrb_string_p(idx)) {
-    mrb_value sym = mrb_check_intern_str(mrb, idx);
-
-    if (mrb_nil_p(sym)) {
-      mrb_name_error(mrb, mrb_intern_str(mrb, idx), "no member '%S' in struct", idx);
-    }
-    idx = sym;
+    mrb_sym sym = mrb_intern_str(mrb, idx);
+    idx = mrb_symbol_value(sym);
   }
   if (mrb_symbol_p(idx)) {
     return struct_aref_sym(mrb, s, mrb_symbol(idx));
   }
-  return struct_aref_int(mrb, s, mrb_int(mrb, idx));
+  return struct_aref_int(mrb, s, mrb_as_int(mrb, idx));
 }
 
 static mrb_value
 mrb_struct_aset_sym(mrb_state *mrb, mrb_value s, mrb_sym id, mrb_value val)
 {
-  mrb_value members, *ptr;
-  const mrb_value *ptr_members;
-  mrb_int i, len;
-
-  members = struct_members(mrb, s);
-  len = RARRAY_LEN(members);
-  ptr = RSTRUCT_PTR(s);
-  ptr_members = RARRAY_PTR(members);
-  for (i=0; i<len; i++) {
+  mrb_value members = struct_members(mrb, s);
+  mrb_int len = RARRAY_LEN(members);
+  const mrb_value *ptr_members = RARRAY_PTR(members);
+  for (mrb_int i=0; i<len; i++) {
     if (mrb_symbol(ptr_members[i]) == id) {
-      mrb_struct_modify(mrb, s);
-      ptr[i] = val;
+      mrb_ary_set(mrb, s, i, val);
       return val;
     }
   }
-  mrb_name_error(mrb, id, "no member '%S' in struct", mrb_sym2str(mrb, id));
+  mrb_name_error(mrb, id, "no member '%n' in struct", id);
   return val;                   /* not reach */
 }
 
@@ -504,38 +461,22 @@ mrb_struct_aset_sym(mrb_state *mrb, mrb_value s, mrb_sym id, mrb_value val)
 static mrb_value
 mrb_struct_aset(mrb_state *mrb, mrb_value s)
 {
-  mrb_int i;
   mrb_value idx;
   mrb_value val;
 
   mrb_get_args(mrb, "oo", &idx, &val);
 
   if (mrb_string_p(idx)) {
-    mrb_value sym = mrb_check_intern_str(mrb, idx);
-
-    if (mrb_nil_p(sym)) {
-      mrb_name_error(mrb, mrb_intern_str(mrb, idx), "no member '%S' in struct", idx);
-    }
-    idx = sym;
+    mrb_sym sym = mrb_intern_str(mrb, idx);
+    idx = mrb_symbol_value(sym);
   }
   if (mrb_symbol_p(idx)) {
     return mrb_struct_aset_sym(mrb, s, mrb_symbol(idx), val);
   }
 
-  i = mrb_int(mrb, idx);
-  if (i < 0) i = RSTRUCT_LEN(s) + i;
-  if (i < 0) {
-    mrb_raisef(mrb, E_INDEX_ERROR,
-               "offset %S too small for struct(size:%S)",
-               mrb_fixnum_value(i), mrb_fixnum_value(RSTRUCT_LEN(s)));
-  }
-  if (RSTRUCT_LEN(s) <= i) {
-    mrb_raisef(mrb, E_INDEX_ERROR,
-               "offset %S too large for struct(size:%S)",
-               mrb_fixnum_value(i), mrb_fixnum_value(RSTRUCT_LEN(s)));
-  }
-  mrb_struct_modify(mrb, s);
-  return RSTRUCT_PTR(s)[i] = val;
+  mrb_int i = struct_index(mrb, mrb_as_int(mrb, idx), num_members(mrb, s));
+  mrb_ary_set(mrb, s, i, val);
+  return val;
 }
 
 /* 15.2.18.4.1  */
@@ -559,11 +500,8 @@ mrb_struct_aset(mrb_state *mrb, mrb_value s)
 static mrb_value
 mrb_struct_equal(mrb_state *mrb, mrb_value s)
 {
-  mrb_value s2;
-  mrb_value *ptr, *ptr2;
-  mrb_int i, len;
+  mrb_value s2 = mrb_get_arg1(mrb);
 
-  mrb_get_args(mrb, "o", &s2);
   if (mrb_obj_equal(mrb, s, s2)) {
     return mrb_true_value();
   }
@@ -571,15 +509,17 @@ mrb_struct_equal(mrb_state *mrb, mrb_value s)
     return mrb_false_value();
   }
   if (RSTRUCT_LEN(s) != RSTRUCT_LEN(s2)) {
-    mrb_bug(mrb, "inconsistent struct"); /* should never happen */
+    return mrb_false_value();
   }
-  ptr = RSTRUCT_PTR(s);
-  ptr2 = RSTRUCT_PTR(s2);
-  len = RSTRUCT_LEN(s);
-  for (i=0; i<len; i++) {
+  mrb_value *ptr = RSTRUCT_PTR(s);
+  mrb_value *ptr2 = RSTRUCT_PTR(s2);
+  mrb_int len = RSTRUCT_LEN(s);
+  int ai = mrb_gc_arena_save(mrb);
+  for (mrb_int i=0; i<len; i++) {
     if (!mrb_equal(mrb, ptr[i], ptr2[i])) {
       return mrb_false_value();
     }
+    mrb_gc_arena_restore(mrb, ai);
   }
 
   return mrb_true_value();
@@ -587,7 +527,7 @@ mrb_struct_equal(mrb_state *mrb, mrb_value s)
 
 /* 15.2.18.4.12(x)  */
 /*
- * code-seq:
+ * call-seq:
  *   struct.eql?(other)   -> true or false
  *
  * Two structures are equal if they are the same object, or if all their
@@ -596,11 +536,10 @@ mrb_struct_equal(mrb_state *mrb, mrb_value s)
 static mrb_value
 mrb_struct_eql(mrb_state *mrb, mrb_value s)
 {
-  mrb_value s2;
+  mrb_value s2 = mrb_get_arg1(mrb);
   mrb_value *ptr, *ptr2;
   mrb_int i, len;
 
-  mrb_get_args(mrb, "o", &s2);
   if (mrb_obj_equal(mrb, s, s2)) {
     return mrb_true_value();
   }
@@ -608,7 +547,7 @@ mrb_struct_eql(mrb_state *mrb, mrb_value s)
     return mrb_false_value();
   }
   if (RSTRUCT_LEN(s) != RSTRUCT_LEN(s2)) {
-    mrb_bug(mrb, "inconsistent struct"); /* should never happen */
+    return mrb_false_value();
   }
   ptr = RSTRUCT_PTR(s);
   ptr2 = RSTRUCT_PTR(s2);
@@ -624,8 +563,8 @@ mrb_struct_eql(mrb_state *mrb, mrb_value s)
 
 /*
  * call-seq:
- *    struct.length   -> Fixnum
- *    struct.size     -> Fixnum
+ *    struct.length   -> Integer
+ *    struct.size     -> Integer
  *
  * Returns number of struct members.
  */
@@ -657,14 +596,11 @@ mrb_struct_to_a(mrb_state *mrb, mrb_value self)
 static mrb_value
 mrb_struct_to_h(mrb_state *mrb, mrb_value self)
 {
-  mrb_value members, ret;
-  mrb_int i;
+  mrb_value members = struct_members(mrb, self);
+  mrb_value ret = mrb_hash_new_capa(mrb, RARRAY_LEN(members));
 
-  members = struct_members(mrb, self);
-  ret = mrb_hash_new_capa(mrb, RARRAY_LEN(members));
-
-  for (i = 0; i < RARRAY_LEN(members); ++i) {
-    mrb_hash_set(mrb, ret, RARRAY_PTR(members)[i], RSTRUCT_PTR(self)[i]);
+  for (mrb_int i = 0; i < RARRAY_LEN(members); i++) {
+    mrb_hash_set(mrb, ret, RARRAY_PTR(members)[i], mrb_ary_ref(mrb, self, i));
   }
 
   return ret;
@@ -674,13 +610,52 @@ static mrb_value
 mrb_struct_values_at(mrb_state *mrb, mrb_value self)
 {
   mrb_int argc;
-  mrb_value *argv;
+  const mrb_value *argv;
 
   mrb_get_args(mrb, "*", &argv, &argc);
 
   return mrb_get_values_at(mrb, self, RSTRUCT_LEN(self), argc, argv, struct_aref_int);
 }
 
+/*
+ * call-seq:
+ *    struct.to_s    -> string
+ *    struct.inspect -> string
+ *
+ * Returns a string representation of Data
+ */
+static mrb_value
+mrb_struct_to_s(mrb_state *mrb, mrb_value self)
+{
+  mrb->c->ci->mid = MRB_SYM(inspect);
+  mrb_value ret = mrb_str_new_lit(mrb, "#<struct ");
+  int ai = mrb_gc_arena_save(mrb);
+  mrb_value cname = mrb_class_path(mrb, mrb_class_real(mrb_class(mrb, self)));
+  if (!mrb_nil_p(cname)) {
+    mrb_str_cat_str(mrb, ret, cname);
+    mrb_str_cat_lit(mrb, ret, " ");
+  }
+  if (mrb_inspect_recursive_p(mrb, self)) {
+    mrb_str_cat_lit(mrb, ret, "...>");
+    return ret;
+  }
+  mrb_value members = struct_members(mrb, self);
+  mrb_int mlen = RARRAY_LEN(members);
+  mrb_value *mems = RARRAY_PTR(members);
+  for (mrb_int i=0; i<mlen; i++) {
+    mrb_int len;
+    const char *name = mrb_sym_name_len(mrb, mrb_symbol(mems[i]), &len);
+    if (i>0) mrb_str_cat_lit(mrb, ret, ", ");
+    mrb_str_cat(mrb, ret, name, len);
+    mrb_str_cat_lit(mrb, ret, "=");
+    mrb_str_cat_str(mrb, ret, mrb_inspect(mrb, mrb_ary_ref(mrb, self, i)));
+    mrb_gc_arena_restore(mrb, ai);
+  }
+  mrb_str_cat_lit(mrb, ret, ">");
+
+  return ret;
+}
+
 /*
  *  A <code>Struct</code> is a convenient way to bundle a number of
  *  attributes together, using accessor methods, without having to write
@@ -699,26 +674,28 @@ mrb_struct_values_at(mrb_state *mrb, mrb_value self)
 void
 mrb_mruby_struct_gem_init(mrb_state* mrb)
 {
-  struct RClass *st;
-  st = mrb_define_class(mrb, "Struct",  mrb->object_class);
-  MRB_SET_INSTANCE_TT(st, MRB_TT_ARRAY);
+  struct RClass *st = mrb_define_class_id(mrb, MRB_SYM(Struct),  mrb->object_class);
+  MRB_SET_INSTANCE_TT(st, MRB_TT_STRUCT);
+  MRB_UNDEF_ALLOCATOR(st);
 
-  mrb_define_class_method(mrb, st, "new",             mrb_struct_s_def,       MRB_ARGS_ANY());  /* 15.2.18.3.1  */
+  mrb_define_class_method_id(mrb, st, MRB_SYM(new), mrb_struct_s_def, MRB_ARGS_ANY());  /* 15.2.18.3.1  */
 
-  mrb_define_method(mrb, st,       "==",              mrb_struct_equal,       MRB_ARGS_REQ(1)); /* 15.2.18.4.1  */
-  mrb_define_method(mrb, st,       "[]",              mrb_struct_aref,        MRB_ARGS_REQ(1)); /* 15.2.18.4.2  */
-  mrb_define_method(mrb, st,       "[]=",             mrb_struct_aset,        MRB_ARGS_REQ(2)); /* 15.2.18.4.3  */
-  mrb_define_method(mrb, st,       "members",         mrb_struct_members,     MRB_ARGS_NONE()); /* 15.2.18.4.6  */
-  mrb_define_method(mrb, st,       "initialize",      mrb_struct_initialize,  MRB_ARGS_ANY());  /* 15.2.18.4.8  */
-  mrb_define_method(mrb, st,       "initialize_copy", mrb_struct_init_copy,   MRB_ARGS_REQ(1)); /* 15.2.18.4.9  */
-  mrb_define_method(mrb, st,       "eql?",            mrb_struct_eql,         MRB_ARGS_REQ(1)); /* 15.2.18.4.12(x)  */
+  mrb_define_method_id(mrb, st, MRB_OPSYM(eq),            mrb_struct_equal,       MRB_ARGS_REQ(1)); /* 15.2.18.4.1  */
+  mrb_define_method_id(mrb, st, MRB_OPSYM(aref),          mrb_struct_aref,        MRB_ARGS_REQ(1)); /* 15.2.18.4.2  */
+  mrb_define_method_id(mrb, st, MRB_OPSYM(aset),          mrb_struct_aset,        MRB_ARGS_REQ(2)); /* 15.2.18.4.3  */
+  mrb_define_method_id(mrb, st, MRB_SYM(members),         mrb_struct_members,     MRB_ARGS_NONE()); /* 15.2.18.4.6  */
+  mrb_define_method_id(mrb, st, MRB_SYM(initialize),      mrb_struct_initialize,  MRB_ARGS_ANY());  /* 15.2.18.4.8  */
+  mrb_define_method_id(mrb, st, MRB_SYM(initialize_copy), mrb_struct_init_copy,   MRB_ARGS_REQ(1)); /* 15.2.18.4.9  */
+  mrb_define_method_id(mrb, st, MRB_SYM_Q(eql),           mrb_struct_eql,         MRB_ARGS_REQ(1)); /* 15.2.18.4.12(x)  */
+  mrb_define_method_id(mrb, st, MRB_SYM(to_s),            mrb_struct_to_s,        MRB_ARGS_NONE()); /* 15.2.18.4.11(x) */
+  mrb_define_method_id(mrb, st, MRB_SYM(inspect),         mrb_struct_to_s,        MRB_ARGS_NONE()); /* 15.2.18.4.10(x) */
 
-  mrb_define_method(mrb, st,        "size",           mrb_struct_len,         MRB_ARGS_NONE());
-  mrb_define_method(mrb, st,        "length",         mrb_struct_len,         MRB_ARGS_NONE());
-  mrb_define_method(mrb, st,        "to_a",           mrb_struct_to_a,        MRB_ARGS_NONE());
-  mrb_define_method(mrb, st,        "values",         mrb_struct_to_a,        MRB_ARGS_NONE());
-  mrb_define_method(mrb, st,        "to_h",           mrb_struct_to_h,        MRB_ARGS_NONE());
-  mrb_define_method(mrb, st,        "values_at",      mrb_struct_values_at,   MRB_ARGS_ANY());
+  mrb_define_method_id(mrb, st, MRB_SYM(size),            mrb_struct_len,         MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, st, MRB_SYM(length),          mrb_struct_len,         MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, st, MRB_SYM(to_a),            mrb_struct_to_a,        MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, st, MRB_SYM(values),          mrb_struct_to_a,        MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, st, MRB_SYM(to_h),            mrb_struct_to_h,        MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, st, MRB_SYM(values_at),       mrb_struct_values_at,   MRB_ARGS_ANY());
 }
 
 void
