@@ -119,7 +119,7 @@ assert('struct inspect') do
   c = Struct.new(:m1, :m2, :m3, :m4, :m5, :recur)
   cc = c.new(1,2,3,4,5,nil)
   cc.recur = cc
-  assert_equal "#<struct m1=1, m2=2, m3=3, m4=4, m5=5, recur=#<struct #{cc.class}:...>>", cc.inspect
+  assert_equal "#<struct m1=1, m2=2, m3=3, m4=4, m5=5, recur=#<struct ...>>", cc.inspect
 end
 
 assert('Struct#length, Struct#size') do
@@ -136,7 +136,7 @@ end
 
 assert('Struct#to_h') do
   s = Struct.new(:white, :red, :green).new('ruuko', 'yuzuki', 'hitoe')
-  assert_equal(:white => 'ruuko', :red => 'yuzuki', :green => 'hitoe') { s.to_h }
+  assert_equal({:white => 'ruuko', :red => 'yuzuki', :green => 'hitoe'}) { s.to_h }
 end
 
 assert('Struct#values_at') do
@@ -166,13 +166,13 @@ assert("Struct#initialize_copy requires struct to be the same type") do
   begin
     Struct.new("Test", :a)
     a = Struct::Test.new("a")
-    Struct.remove_const :Test
+    Struct.__send__(:remove_const,:Test)
     Struct.new("Test", :a, :b)
     assert_raise(TypeError) do
       a.initialize_copy(Struct::Test.new("a", "b"))
     end
   ensure
-    Struct.remove_const :Test
+    Struct.__send__(:remove_const,:Test)
   end
 end
 
