@@ -89,10 +89,7 @@ print_args(uint16_t i, FILE *out)
 static void
 codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
 {
-  int ai;
-  const mrb_code *pc, *pcend;
-  mrb_code ins;
-  const char *file = NULL, *next_file;
+  const char *file = NULL;
 
   if (!irep) return;
   fprintf(out, "irep %p nregs=%d nlocals=%d pools=%d syms=%d reps=%d ilen=%d\n", (void*)irep,
@@ -140,18 +137,18 @@ codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
     }
   }
 
-  pc = irep->iseq;
-  pcend = pc + irep->ilen;
+  const mrb_code *pc = irep->iseq;
+  const mrb_code *pcend = pc + irep->ilen;
   while (pc < pcend) {
-    ptrdiff_t i;
     uint32_t a;
     uint16_t b;
     uint16_t c;
+    mrb_code ins;
 
-    ai = mrb_gc_arena_save(mrb);
+    int ai = mrb_gc_arena_save(mrb);
+    ptrdiff_t i = pc - irep->iseq;
 
-    i = pc - irep->iseq;
-    next_file = mrb_debug_get_filename(mrb, irep, (uint32_t)i);
+    const char *next_file = mrb_debug_get_filename(mrb, irep, (uint32_t)i);
     if (next_file && file != next_file) {
       fprintf(out, "file: %s\n", next_file);
       file = next_file;
@@ -188,8 +185,8 @@ codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
       }
       print_lv_a(mrb, irep, a, out);
       break;
-    CASE(OP_LOADI, BB):
-      fprintf(out, "LOADI\t\tR%d\t%d\t", a, b);
+    CASE(OP_LOADI8, BB):
+      fprintf(out, "LOADI8\tR%d\t%d\t", a, b);
       print_lv_a(mrb, irep, a, out);
       break;
     CASE(OP_LOADINEG, BB):
@@ -416,7 +413,7 @@ codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
       fprintf(out, "RANGE_EXC\tR%d\n", a);
       break;
     CASE(OP_DEF, BB):
-      fprintf(out, "DEF\t\tR%d\t:%s\n", a, mrb_sym_dump(mrb, irep->syms[b]));
+      fprintf(out, "DEF\t\tR%d\t:%s\t(R%d)\n", a, mrb_sym_dump(mrb, irep->syms[b]),a+1);
       break;
     CASE(OP_UNDEF, B):
       fprintf(out, "UNDEF\t\t:%s\n", mrb_sym_dump(mrb, irep->syms[a]));
@@ -581,7 +578,7 @@ codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
       ins = READ_B();
       switch (ins) {
 #define OPCODE(i,x) case OP_ ## i: FETCH_ ## x ## _1 (); goto L_OP_ ## i;
-#include "mruby/ops.h"
+#include <mruby/ops.h>
 #undef OPCODE
       }
       break;
@@ -591,7 +588,7 @@ codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
       ins = READ_B();
       switch (ins) {
 #define OPCODE(i,x) case OP_ ## i: FETCH_ ## x ## _2 (); goto L_OP_ ## i;
-#include "mruby/ops.h"
+#include <mruby/ops.h>
 #undef OPCODE
       }
       break;
@@ -601,7 +598,7 @@ codedump(mrb_state *mrb, const mrb_irep *irep, FILE *out)
       ins = READ_B();
       switch (ins) {
 #define OPCODE(i,x) case OP_ ## i: FETCH_ ## x ## _3 (); goto L_OP_ ## i;
-#include "mruby/ops.h"
+#include <mruby/ops.h>
 #undef OPCODE
       }
       break;
