@@ -9,13 +9,13 @@
 #include <mruby/data.h>
 #include <mruby/class.h>
 #include <mruby/numeric.h>
+#include <mruby/internal.h>
 
 MRB_API struct RData*
 mrb_data_object_alloc(mrb_state *mrb, struct RClass *klass, void *ptr, const mrb_data_type *type)
 {
-  struct RData *data;
+  struct RData *data = MRB_OBJ_ALLOC(mrb, MRB_TT_CDATA, klass);
 
-  data = MRB_OBJ_ALLOC(mrb, MRB_TT_DATA, klass);
   data->data = ptr;
   data->type = type;
 
@@ -26,7 +26,7 @@ MRB_API void
 mrb_data_check_type(mrb_state *mrb, mrb_value obj, const mrb_data_type *type)
 {
   if (!mrb_data_p(obj)) {
-    mrb_check_type(mrb, obj, MRB_TT_DATA);
+    mrb_check_type(mrb, obj, MRB_TT_CDATA);
   }
   if (DATA_TYPE(obj) != type) {
     const mrb_data_type *t2 = DATA_TYPE(obj);
@@ -70,33 +70,13 @@ mrb_obj_to_sym(mrb_state *mrb, mrb_value name)
   return 0;  /* not reached */
 }
 
+#if !defined(MRB_NO_FLOAT) && !defined(MRB_NAN_BOXING)
 static mrb_int
-make_num_id(const char *p, size_t len)
-{
-  uint32_t id = 0;
-
-  while (len--) {
-    id = id*65599 + *p;
-    p++;
-  }
-  id = id + (id>>5);
-
-  return (mrb_int)id;
-}
-
-MRB_API mrb_int
-mrb_int_id(mrb_int n)
-{
-  return make_num_id((const char*)&n, sizeof(n));
-}
-
-#ifndef MRB_NO_FLOAT
-MRB_API mrb_int
 mrb_float_id(mrb_float f)
 {
   /* normalize -0.0 to 0.0 */
   if (f == 0) f = 0.0;
-  return make_num_id((const char*)&f, sizeof(f));
+  return (mrb_int)mrb_byte_hash((uint8_t*)&f, sizeof(f));
 }
 #endif
 
@@ -140,7 +120,7 @@ mrb_obj_id(mrb_value obj)
   case MRB_TT_SYMBOL:
     return MakeID(mrb_symbol(obj), tt);
   case MRB_TT_INTEGER:
-    return MakeID(mrb_int_id(mrb_integer(obj)), tt);
+    return MakeID(mrb_integer(obj), tt);
 #ifndef MRB_NO_FLOAT
   case MRB_TT_FLOAT:
     return MakeID(mrb_float_id(mrb_float(obj)), tt);
@@ -156,7 +136,7 @@ mrb_obj_id(mrb_value obj)
   case MRB_TT_HASH:
   case MRB_TT_RANGE:
   case MRB_TT_EXCEPTION:
-  case MRB_TT_DATA:
+  case MRB_TT_CDATA:
   case MRB_TT_ISTRUCT:
   default:
     return MakeID(mrb_ptr(obj), tt);
@@ -174,11 +154,11 @@ mrb_word_boxing_float_value(mrb_state *mrb, mrb_float f)
 #ifdef MRB_WORDBOX_NO_FLOAT_TRUNCATE
   v.p = mrb_obj_alloc(mrb, MRB_TT_FLOAT, mrb->float_class);
   v.fp->f = f;
-  MRB_SET_FROZEN_FLAG(v.bp);
+  v.bp->frozen = 1;
 #elif defined(MRB_64BIT) && defined(MRB_USE_FLOAT32)
   v.w = 0;
   v.f = f;
-  v.w = ((v.w<<2) & ~3) | 2;
+  v.w = (v.w<<2) | 2;
 #else
   v.f = f;
   v.w = (v.w & ~3) | 2;
@@ -193,9 +173,10 @@ mrb_word_boxing_value_float(mrb_value v)
 {
   union mrb_value_ u;
   u.value = v;
-  u.w = u.w & ~3;
 #if defined(MRB_64BIT) && defined(MRB_USE_FLOAT32)
   u.w >>= 2;
+#else
+  u.w &= ~3;
 #endif
   return u.f;
 }
@@ -221,11 +202,9 @@ mrb_boxing_int_value(mrb_state *mrb, mrb_int n)
   if (FIXABLE(n)) return mrb_fixnum_value(n);
   else {
     mrb_value v;
-    struct RInteger *p;
-
-    p = (struct RInteger*)mrb_obj_alloc(mrb, MRB_TT_INTEGER, mrb->integer_class);
+    struct RInteger *p = (struct RInteger*)mrb_obj_alloc(mrb, MRB_TT_INTEGER, mrb->integer_class);
     p->i = n;
-    MRB_SET_FROZEN_FLAG((struct RBasic*)p);
+    p->frozen = 1;
     SET_OBJ_VALUE(v, p);
     return v;
   }
@@ -260,9 +239,9 @@ MRB_API int
 mrb_msvc_snprintf(char *s, size_t n, const char *format, ...)
 {
   va_list arg;
-  int ret;
   va_start(arg, format);
-  ret = mrb_msvc_vsnprintf(s, n, format, arg);
+
+  int ret = mrb_msvc_vsnprintf(s, n, format, arg);
   va_end(arg);
   return ret;
 }
