@@ -11,6 +11,7 @@
 #include <mruby/hash.h>
 #include <mruby/string.h>
 #include <mruby/variable.h>
+#include <mruby/proc.h>
 #include <mruby/internal.h>
 #include <mruby/presym.h>
 
@@ -145,78 +146,57 @@ typedef struct index_buckets_iter {
   }
 
 #ifdef MRB_64BIT
-DEFINE_ACCESSOR(ar, ea_capa, uint32_t, ea_capa)
-DEFINE_ACCESSOR(ar, ea_n_used, uint32_t, ea_n_used)
-DEFINE_ACCESSOR(ht, ea_capa, uint32_t, ea_capa)
-DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, ea_n_used)
+DEFINE_ACCESSOR(ar, ea_capa, uint32_t, ea_capa)                 /* ar_ea_capa   ar_set_ea_capa   */
+DEFINE_ACCESSOR(ar, ea_n_used, uint32_t, ea_n_used)             /* ar_ea_n_used ar_set_ea_n_used */
+DEFINE_ACCESSOR(ht, ea_capa, uint32_t, ea_capa)                 /* ht_ea_capa   ht_set_ea_capa   */
+DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, ea_n_used)             /* ht_ea_n_used ht_set_ea_n_used */
 #else
-DEFINE_FLAG_ACCESSOR(ar, ea_capa, uint32_t, AR_EA_CAPA)
-DEFINE_FLAG_ACCESSOR(ar, ea_n_used, uint32_t, AR_EA_N_USED)
-DEFINE_ACCESSOR(ht, ea_capa, uint32_t, hsh.ht->ea_capa)
-DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, hsh.ht->ea_n_used)
+DEFINE_FLAG_ACCESSOR(ar, ea_capa, uint32_t, AR_EA_CAPA)         /* ar_ea_capa   ar_set_ea_capa   */
+DEFINE_FLAG_ACCESSOR(ar, ea_n_used, uint32_t, AR_EA_N_USED)     /* ar_ea_n_used ar_set_ea_n_used */
+DEFINE_ACCESSOR(ht, ea_capa, uint32_t, hsh.ht->ea_capa)         /* ht_ea_capa   ht_set_ea_capa   */
+DEFINE_ACCESSOR(ht, ea_n_used, uint32_t, hsh.ht->ea_n_used)     /* ht_ea_n_used ht_set_ea_n_used */
 #endif
-DEFINE_FLAG_ACCESSOR(ib, bit, uint32_t, IB_BIT)
-DEFINE_ACCESSOR(ar, size, uint32_t, size)
-DEFINE_ACCESSOR(ar, ea, hash_entry*, hsh.ea)
-DEFINE_DECREMENTER(ar, size)
-DEFINE_ACCESSOR(ht, size, uint32_t, size)
-DEFINE_ACCESSOR(ht, ea, hash_entry*, hsh.ht->ea)
-DEFINE_GETTER(ht, ib, uint32_t*, hsh.ht->ib)
-DEFINE_INCREMENTER(ht, size)
-DEFINE_DECREMENTER(ht, size)
-DEFINE_GETTER(h, size, uint32_t, size)
-DEFINE_ACCESSOR(h, ht, hash_table*, hsh.ht)
-DEFINE_SWITCHER(ht, HT)
+DEFINE_FLAG_ACCESSOR(ib, bit, uint32_t, IB_BIT)                 /* ib_bit       ib_set_bit       */
+DEFINE_ACCESSOR(ar, size, uint32_t, size)                       /* ar_size      ar_set_size      */
+DEFINE_ACCESSOR(ar, ea, hash_entry*, hsh.ea)                    /* ar_ea        ar_set_ea        */
+DEFINE_DECREMENTER(ar, size)                                    /* ar_dec_size                   */
+DEFINE_ACCESSOR(ht, size, uint32_t, size)                       /* ht_size      ht_set_size      */
+DEFINE_ACCESSOR(ht, ea, hash_entry*, hsh.ht->ea)                /* ht_ea        ht_set_ea        */
+DEFINE_GETTER(ht, ib, uint32_t*, hsh.ht->ib)                    /* ht_ib                         */
+DEFINE_INCREMENTER(ht, size)                                    /* ht_inc_size                   */
+DEFINE_DECREMENTER(ht, size)                                    /* ht_dec_size                   */
+DEFINE_GETTER(h, size, uint32_t, size)                          /* h_size                        */
+DEFINE_ACCESSOR(h, ht, hash_table*, hsh.ht)                     /* h_ht         h_set_ht         */
+DEFINE_SWITCHER(ht, HT)                                         /* h_ht_on  h_ht_off  h_ht_p     */
 
-#define ea_each_used(ea, n_used, entry_var, code) do {                        \
-  hash_entry *entry_var = ea, *ea_end__ = entry_var + (n_used);               \
-  for (; entry_var < ea_end__; entry_var++) {                                 \
-    code;                                                                     \
-  }                                                                           \
-} while (0)
+#define EA_EACH_USED(ea, n_used, entry_var)                                   \
+  for (hash_entry *entry_var = (ea), *ea_end__ = (entry_var) + (n_used);      \
+       entry_var < ea_end__;                                                  \
+       entry_var++)
 
-#define ea_each(ea, size, entry_var, code) do {                               \
-  hash_entry *entry_var = ea;                                                 \
-  uint32_t size__ = size;                                                     \
-  for (; 0 < size__; entry_var++) {                                           \
-    if (entry_deleted_p(entry_var)) continue;                                 \
-    --size__;                                                                 \
-    code;                                                                     \
-  }                                                                           \
-} while (0)
+#define EA_EACH(ea, size, entry_var)                                          \
+  for (uint32_t ea_size__ = (size); ea_size__; ea_size__ = 0)                 \
+    for (hash_entry *entry_var = (ea);                                        \
+         ea_size__ && (entry_var = entry_skip_deleted(entry_var), TRUE);      \
+         entry_var++, ea_size__--)
 
-#define ib_cycle_by_key(mrb, h, key, it_var, code) do {                       \
-  index_buckets_iter it_var[1];                                               \
-  ib_it_init(mrb, it_var, h, key);                                            \
-  for (;;) {                                                                  \
-    ib_it_next(it_var);                                                       \
-    code;                                                                     \
-  }                                                                           \
-} while (0)
+#define IB_CYCLE_BY_KEY(mrb, h, key, it_var)                                  \
+  for (index_buckets_iter it_var[1] = { ib_it_init(mrb, h, key) };            \
+       (ib_it_next(it_var), TRUE);                                            \
+       /* do nothing */)
 
-#define ib_find_by_key(mrb, h_, key_, it_var, code) do {                      \
-  mrb_value ib_fbk_key__ = key_;                                              \
-  ib_cycle_by_key(mrb, h_, ib_fbk_key__, it_var, {                            \
-    if (ib_it_empty_p(it_var)) break;                                         \
-    if (ib_it_deleted_p(it_var)) continue;                                    \
-    if (obj_eql(mrb, ib_fbk_key__, ib_it_entry(it_var)->key, it_var->h)) {    \
-      code;                                                                   \
-      break;                                                                  \
-    }                                                                         \
-  });                                                                         \
-} while (0)
+#define IB_FIND_BY_KEY(mrb, h, key, it_var)                                   \
+  for (index_buckets_iter it_var[1] = { ib_it_init(mrb, h, key) };            \
+       ib_it_find_by_key(mrb, it_var, key);                                   \
+       it_var[0].h = NULL)
 
-#define h_each(h, entry_var, code) do {                                       \
-  struct RHash *h__ = h;                                                      \
-  hash_entry *h_e_ea__;                                                       \
-  uint32_t h_e_size__;                                                        \
-  h_ar_p(h) ? (h_e_ea__ = ar_ea(h__), h_e_size__ = ar_size(h__)) :            \
-              (h_e_ea__ = ht_ea(h__), h_e_size__ = ht_size(h__));             \
-  ea_each(h_e_ea__, h_e_size__, entry_var, code);                             \
-} while (0)
+#define H_EACH(h, entry_var)                                                  \
+  EA_EACH((h_ar_p(h) ? ar_ea(h) : ht_ea(h)),                                  \
+          (h_ar_p(h) ? ar_size(h) : ht_size(h)),                              \
+          entry_var)
 
 /*
- * In `h_check_modified()`, in the case of `MRB_NO_BOXING`, `ht_ea()` or
+ * In `H_CHECK_MODIFIED()`, in the case of `MRB_NO_BOXING`, `ht_ea()` or
  * `ht_ea_capa()` for AR may read uninitialized area (#5332). Therefore, do
  * not use those macros for AR in `MRB_NO_BOXING` (but in the case of
  * `MRB_64BIT`, `ht_ea_capa()` is the same as `ar_ea_capa()`, so use it).
@@ -232,10 +212,10 @@ DEFINE_SWITCHER(ht, HT)
 # define H_CHECK_MODIFIED_USE_HT_EA_FOR_AR TRUE
 # define H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR TRUE
  /*
-  * `h_check_modified` raises an exception when a dangerous modification is
+  * `H_CHECK_MODIFIED` raises an exception when a dangerous modification is
   * made to `h` by executing `code`.
   *
-  * `h_check_modified` macro is not called if `h->hsh.ht` (`h->hsh.ea`) is `NULL`
+  * `H_CHECK_MODIFIED` macro is not called if `h->hsh.ht` (`h->hsh.ea`) is `NULL`
   * (`Hash` size is zero). And because the `hash_entry` is rather large,
   * `h->hsh.ht->ea` and `h->hsh.ht->ea_capa` are able to be safely accessed even for
   * AR. This nature is used to eliminate branch of AR or HT.
@@ -255,32 +235,13 @@ HT_ASSERT_SAFE_READ(ea_capa);
 #endif  /* MRB_NO_BOXING */
 
 /*
- * `h_check_modified` raises an exception when a dangerous modification is
- * made to `h` by executing `code`.
+ * `H_CHECK_MODIFIED` raises an exception when a dangerous modification is
+ * made to `h` by executing code block.
  */
-#define h_check_modified(mrb, h, code) do {                                     \
-  struct RHash *h__ = h;                                                        \
-  uint32_t mask__ = MRB_HASH_HT|MRB_HASH_IB_BIT_MASK|MRB_HASH_AR_EA_CAPA_MASK;  \
-  uint32_t flags__ = h__->flags & mask__;                                       \
-  void* tbl__ = (mrb_assert(h__->hsh.ht), h__->hsh.ht);                         \
-  uint32_t ht_ea_capa__ = 0;                                                    \
-  hash_entry *ht_ea__ = NULL;                                                   \
-  if (H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR || h_ht_p(h__)) {                  \
-    ht_ea_capa__ = ht_ea_capa(h__);                                             \
-  }                                                                             \
-  if (H_CHECK_MODIFIED_USE_HT_EA_FOR_AR || h_ht_p(h__)) {                       \
-    ht_ea__ = ht_ea(h__);                                                       \
-  }                                                                             \
-  code;                                                                         \
-  if (flags__ != (h__->flags & mask__) ||                                       \
-      tbl__ != h__->hsh.ht ||                                                       \
-      ((H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR || h_ht_p(h__)) &&               \
-       ht_ea_capa__ != ht_ea_capa(h__)) ||                                      \
-      ((H_CHECK_MODIFIED_USE_HT_EA_FOR_AR || h_ht_p(h__)) &&                    \
-       ht_ea__ != ht_ea(h__))) {                                                \
-    mrb_raise(mrb, E_RUNTIME_ERROR, "hash modified");                           \
-  }                                                                             \
-} while (0)
+#define H_CHECK_MODIFIED(mrb, h)                                              \
+  for (struct h_check_modified h_checker__ = h_check_modified_init(mrb, h);   \
+       h_checker__.tbl;                                                       \
+       h_check_modified_validate(mrb, &h_checker__, h), h_checker__.tbl = NULL)
 
 #define U32(v) ((uint32_t)(v))
 #define h_ar_p(h) (!h_ht_p(h))
@@ -291,12 +252,11 @@ HT_ASSERT_SAFE_READ(ea_capa);
 
 static uint32_t ib_upper_bound_for(uint32_t capa);
 static uint32_t ib_bit_to_capa(uint32_t bit);
+static hash_entry *ib_it_entry(index_buckets_iter *it);
 static void ht_init(
   mrb_state *mrb, struct RHash *h, uint32_t size,
   hash_entry *ea, uint32_t ea_capa, hash_table *ht, uint32_t ib_bit);
 static void ht_set(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val);
-static void ht_set_without_ib_adjustment(
-  mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val);
 
 static uint32_t
 next_power2(uint32_t v)
@@ -315,6 +275,41 @@ next_power2(uint32_t v)
 #endif
 }
 
+struct h_check_modified {
+  uint32_t flags;
+  void *tbl;
+  uint32_t ht_ea_capa;
+  hash_entry *ht_ea;
+};
+
+#define H_CHECK_MODIFIED_FLAGS_MASK  (MRB_HASH_HT | MRB_HASH_IB_BIT_MASK | MRB_HASH_AR_EA_CAPA_MASK)
+
+static struct h_check_modified
+h_check_modified_init(mrb_state *mrb, struct RHash *h)
+{
+  mrb_assert(h->hsh.ht);
+
+  struct h_check_modified checker;
+  checker.flags = h->flags & H_CHECK_MODIFIED_FLAGS_MASK;
+  checker.tbl = h->hsh.ht;
+  checker.ht_ea_capa = (H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR || h_ht_p(h)) ? ht_ea_capa(h) : 0;
+  checker.ht_ea = (H_CHECK_MODIFIED_USE_HT_EA_FOR_AR || h_ht_p(h)) ? ht_ea(h) : NULL;
+  return checker;
+}
+
+static void
+h_check_modified_validate(mrb_state *mrb, struct h_check_modified *checker, struct RHash *h)
+{
+  if (checker->flags != (h->flags & H_CHECK_MODIFIED_FLAGS_MASK) ||
+      checker->tbl != h->hsh.ht ||
+      ((H_CHECK_MODIFIED_USE_HT_EA_CAPA_FOR_AR || h_ht_p(h)) &&
+       checker->ht_ea_capa != ht_ea_capa(h)) ||
+      ((H_CHECK_MODIFIED_USE_HT_EA_FOR_AR || h_ht_p(h)) &&
+       checker->ht_ea != ht_ea(h))) {
+    mrb_raise(mrb, E_RUNTIME_ERROR, "hash modified");
+  }
+}
+
 static uint32_t
 obj_hash_code(mrb_state *mrb, mrb_value key, struct RHash *h)
 {
@@ -342,9 +337,9 @@ obj_hash_code(mrb_state *mrb, mrb_value key, struct RHash *h)
     hash_code = U32(mrb_obj_id(key));
     break;
   default:
-    h_check_modified(mrb, h, {
+    H_CHECK_MODIFIED(mrb, h) {
       hash_code_obj = mrb_funcall_argv(mrb, key, MRB_SYM(hash), 0, NULL);
-    });
+    }
 
     hash_code = U32(tt) ^ U32(mrb_integer(hash_code_obj));
     break;
@@ -355,10 +350,9 @@ obj_hash_code(mrb_state *mrb, mrb_value key, struct RHash *h)
 static mrb_bool
 obj_eql(mrb_state *mrb, mrb_value a, mrb_value b, struct RHash *h)
 {
-  enum mrb_vtype tt = mrb_type(a);
   mrb_bool eql;
 
-  switch (tt) {
+  switch (mrb_type(a)) {
   case MRB_TT_STRING:
     return mrb_str_equal(mrb, a, b);
 
@@ -377,12 +371,12 @@ obj_eql(mrb_state *mrb, mrb_value a, mrb_value b, struct RHash *h)
 #endif
 
   default:
-    h_check_modified(mrb, h, {eql = mrb_eql(mrb, a, b);});
+    H_CHECK_MODIFIED(mrb, h) {eql = mrb_eql(mrb, a, b);}
     return eql;
   }
 }
 
-static mrb_bool
+static inline mrb_bool
 entry_deleted_p(const hash_entry* entry)
 {
   return mrb_undef_p(entry->key);
@@ -394,6 +388,14 @@ entry_delete(hash_entry* entry)
   entry->key = mrb_undef_value();
 }
 
+static hash_entry*
+entry_skip_deleted(hash_entry *e)
+{
+  for (; entry_deleted_p(e); e++)
+    ;
+  return e;
+}
+
 static uint32_t
 ea_next_capa_for(uint32_t size, uint32_t max_capa)
 {
@@ -423,11 +425,11 @@ static void
 ea_compress(hash_entry *ea, uint32_t n_used)
 {
   hash_entry *w_entry = ea;
-  ea_each_used(ea, n_used, r_entry, {
+  EA_EACH_USED(ea, n_used, r_entry) {
     if (entry_deleted_p(r_entry)) continue;
     if (r_entry != w_entry) *w_entry = *r_entry;
     w_entry++;
-  });
+  }
 }
 
 /*
@@ -454,9 +456,9 @@ static hash_entry*
 ea_get_by_key(mrb_state *mrb, hash_entry *ea, uint32_t size, mrb_value key,
               struct RHash *h)
 {
-  ea_each(ea, size, entry, {
+  EA_EACH(ea, size, entry) {
     if (obj_eql(mrb, key, entry->key, h)) return entry;
-  });
+  }
   return NULL;
 }
 
@@ -511,11 +513,11 @@ ar_compress(mrb_state *mrb, struct RHash *h)
 static mrb_bool
 ar_get(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value *valp)
 {
-  ea_each(ar_ea(h), ar_size(h), entry, {
+  EA_EACH(ar_ea(h), ar_size(h), entry) {
     if (!obj_eql(mrb, key, entry->key, h)) continue;
     *valp = entry->val;
     return TRUE;
-  });
+  }
   return FALSE;
 }
 
@@ -566,13 +568,13 @@ static void
 ar_shift(mrb_state *mrb, struct RHash *h, mrb_value *keyp, mrb_value *valp)
 {
   uint32_t size = ar_size(h);
-  ea_each(ar_ea(h), size, entry, {
+  EA_EACH(ar_ea(h), size, entry) {
     *keyp = entry->key;
     *valp = entry->val;
     entry_delete(entry);
     ar_set_size(h, --size);
     return;
-  });
+  }
 }
 
 static void
@@ -581,7 +583,7 @@ ar_rehash(mrb_state *mrb, struct RHash *h)
   /* see comments in `h_rehash` */
   uint32_t size = ar_size(h), w_size = 0, ea_capa = ar_ea_capa(h);
   hash_entry *ea = ar_ea(h), *w_entry;
-  ea_each(ea, size, r_entry, {
+  EA_EACH(ea, size, r_entry) {
     if ((w_entry = ea_get_by_key(mrb, ea, w_size, r_entry->key, h))) {
       w_entry->val = r_entry->val;
       ar_set_size(h, --size);
@@ -594,7 +596,7 @@ ar_rehash(mrb_state *mrb, struct RHash *h)
       }
       w_size++;
     }
-  });
+  }
   mrb_assert(size == w_size);
   ar_set_ea_n_used(h, size);
   ar_adjust_ea(mrb, h, size, ea_capa);
@@ -636,14 +638,16 @@ ib_it_active_p(const index_buckets_iter *it)
   return it->ea_index < ib_it_deleted_value(it);
 }
 
-static void
-ib_it_init(mrb_state *mrb, index_buckets_iter *it, struct RHash *h, mrb_value key)
+static index_buckets_iter
+ib_it_init(mrb_state *mrb, struct RHash *h, mrb_value key)
 {
-  it->h = h;
-  it->bit = ib_bit(h);
-  it->mask = ib_bit_to_capa(it->bit) - 1;
-  it->pos = ib_it_pos_for(it, obj_hash_code(mrb, key, h));
-  it->step = 0;
+  index_buckets_iter it;
+  it.h = h;
+  it.bit = ib_bit(h);
+  it.mask = ib_bit_to_capa(it.bit) - 1;
+  it.pos = ib_it_pos_for(&it, obj_hash_code(mrb, key, h));
+  it.step = 0;
+  return it;
 }
 
 static void
@@ -685,6 +689,21 @@ ib_it_next(index_buckets_iter *it)
   it->pos = ib_it_pos_for(it, it->pos + (++it->step));
 }
 
+static mrb_bool
+ib_it_find_by_key(mrb_state *mrb, index_buckets_iter *it, mrb_value key)
+{
+  if (!it->h) return FALSE;
+
+  for (;;) {
+    ib_it_next(it);
+    if (ib_it_empty_p(it)) return FALSE;
+    if (!ib_it_deleted_p(it) &&
+        obj_eql(mrb, key, ib_it_entry(it)->key, it->h)) {
+      return TRUE;
+    }
+  }
+}
+
 static uint32_t
 ib_it_get(const index_buckets_iter *it)
 {
@@ -769,13 +788,13 @@ ib_init(mrb_state *mrb, struct RHash *h, uint32_t ib_bit, size_t ib_byte_size)
   hash_entry *ea = ht_ea(h);
   memset(ht_ib(h), 0xff, ib_byte_size);
   ib_set_bit(h, ib_bit);
-  ea_each_used(ea, ht_ea_n_used(h), entry, {
-    ib_cycle_by_key(mrb, h, entry->key, it, {
+  EA_EACH_USED(ea, ht_ea_n_used(h), entry) {
+    IB_CYCLE_BY_KEY(mrb, h, entry->key, it) {
       if (!ib_it_empty_p(it)) continue;
       ib_it_set(it, U32(entry - ea));
       break;
-    });
-  });
+    }
+  }
 }
 
 static void
@@ -833,10 +852,10 @@ ht_to_ar(mrb_state *mrb, struct RHash *h)
 static mrb_bool
 ht_get(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value *valp)
 {
-  ib_find_by_key(mrb, h, key, it, {
+  IB_FIND_BY_KEY(mrb, h, key, it) {
     *valp = ib_it_entry(it)->val;
     return TRUE;
-  });
+  }
   return FALSE;
 }
 
@@ -847,32 +866,6 @@ ht_set_as_ar(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val)
   ar_set(mrb, h, key, val);
 }
 
-static void
-ht_set_without_ib_adjustment(mrb_state *mrb, struct RHash *h,
-                             mrb_value key, mrb_value val)
-{
-  mrb_assert(ht_size(h) < ib_bit_to_capa(ib_bit(h)));
-  ib_cycle_by_key(mrb, h, key, it, {
-    if (ib_it_active_p(it)) {
-      if (!obj_eql(mrb, key, ib_it_entry(it)->key, h)) continue;
-      ib_it_entry(it)->val = val;
-    }
-    else {
-      uint32_t ea_n_used = ht_ea_n_used(h);
-      if (ea_n_used == H_MAX_SIZE) {
-        mrb_assert(ht_size(h) == ea_n_used);
-        mrb_raise(mrb, E_ARGUMENT_ERROR, "hash too big");
-      }
-      if (ea_n_used == ht_ea_capa(h)) ht_adjust_ea(mrb, h, ea_n_used, EA_MAX_CAPA);
-      ib_it_set(it, ea_n_used);
-      ea_set(ht_ea(h), ea_n_used, key, val);
-      ht_inc_size(h);
-      ht_set_ea_n_used(h, ++ea_n_used);
-    }
-    return;
-  });
-}
-
 static void
 ht_set(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val)
 {
@@ -885,7 +878,10 @@ ht_set(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val)
   else if (size != ht_ea_n_used(h)) {
     if (ib_capa - EA_N_RESERVED_INDICES <= ht_ea_n_used(h)) goto compress;
     if (ht_ea_capa(h) == ht_ea_n_used(h)) {
-      if (size <= AR_MAX_SIZE) {ht_set_as_ar(mrb, h, key, val); return;}
+      if (size <= AR_MAX_SIZE) {
+        ht_set_as_ar(mrb, h, key, val);
+        return;
+      }
       if (ea_next_capa_for(size, EA_MAX_CAPA) <= ht_ea_capa(h)) {
        compress:
         ea_compress(ht_ea(h), ht_ea_n_used(h));
@@ -894,20 +890,43 @@ ht_set(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value val)
       }
     }
   }
-  ht_set_without_ib_adjustment(mrb, h, key, val);
+
+  mrb_assert(ht_size(h) < ib_bit_to_capa(ib_bit(h)));
+  IB_CYCLE_BY_KEY(mrb, h, key, it) {
+    if (ib_it_active_p(it)) {
+      if (!obj_eql(mrb, key, ib_it_entry(it)->key, h)) continue;
+      ib_it_entry(it)->val = val;
+    }
+    else if (ib_it_deleted_p(it)) {
+      continue;
+    }
+    else {
+      uint32_t ea_n_used = ht_ea_n_used(h);
+      if (ea_n_used == H_MAX_SIZE) {
+        mrb_assert(ht_size(h) == ea_n_used);
+        mrb_raise(mrb, E_ARGUMENT_ERROR, "hash too big");
+      }
+      if (ea_n_used == ht_ea_capa(h)) ht_adjust_ea(mrb, h, ea_n_used, EA_MAX_CAPA);
+      ib_it_set(it, ea_n_used);
+      ea_set(ht_ea(h), ea_n_used, key, val);
+      ht_inc_size(h);
+      ht_set_ea_n_used(h, ++ea_n_used);
+    }
+    return;
+  }
 }
 
 static mrb_bool
 ht_delete(mrb_state *mrb, struct RHash *h, mrb_value key, mrb_value *valp)
 {
-  ib_find_by_key(mrb, h, key, it, {
+  IB_FIND_BY_KEY(mrb, h, key, it) {
     hash_entry *entry = ib_it_entry(it);
     *valp = entry->val;
     ib_it_delete(it);
     entry_delete(entry);
     ht_dec_size(h);
     return TRUE;
-  });
+  }
   return FALSE;
 }
 
@@ -915,8 +934,8 @@ static void
 ht_shift(mrb_state *mrb, struct RHash *h, mrb_value *keyp, mrb_value *valp)
 {
   hash_entry *ea = ht_ea(h);
-  ea_each(ea, ht_size(h), entry, {
-    ib_cycle_by_key(mrb, h, entry->key, it, {
+  EA_EACH(ea, ht_size(h), entry) {
+    IB_CYCLE_BY_KEY(mrb, h, entry->key, it) {
       if (ib_it_get(it) != U32(entry - ea)) continue;
       *keyp = entry->key;
       *valp = entry->val;
@@ -924,8 +943,8 @@ ht_shift(mrb_state *mrb, struct RHash *h, mrb_value *keyp, mrb_value *valp)
       entry_delete(entry);
       ht_dec_size(h);
       return;
-    });
-  });
+    }
+  }
 }
 
 static void
@@ -943,8 +962,8 @@ ht_rehash(mrb_state *mrb, struct RHash *h)
   ht_init(mrb, h, 0, ea, ea_capa, h_ht(h), ib_bit_for(size));
   ht_set_size(h, size);
   ht_set_ea_n_used(h, ht_ea_n_used(h));
-  ea_each(ea, size, r_entry, {
-    ib_cycle_by_key(mrb, h, r_entry->key, it, {
+  EA_EACH(ea, size, r_entry) {
+    IB_CYCLE_BY_KEY(mrb, h, r_entry->key, it) {
       if (ib_it_active_p(it)) {
         if (!obj_eql(mrb, r_entry->key, ib_it_entry(it)->key, h)) continue;
         ib_it_entry(it)->val = r_entry->val;
@@ -959,8 +978,8 @@ ht_rehash(mrb_state *mrb, struct RHash *h)
         ib_it_set(it, w_size++);
       }
       break;
-    });
-  });
+    }
+  }
   mrb_assert(size == w_size);
   ht_set_ea_n_used(h, size);
   size <= AR_MAX_SIZE ? ht_to_ar(mrb, h) : ht_adjust_ea(mrb, h, size, ea_capa);
@@ -969,9 +988,9 @@ ht_rehash(mrb_state *mrb, struct RHash *h)
 static mrb_value
 h_key_for(mrb_state *mrb, mrb_value key)
 {
-  if (mrb_string_p(key) && !MRB_FROZEN_P(mrb_str_ptr(key))) {
+  if (mrb_string_p(key) && !mrb_frozen_p(mrb_str_ptr(key))) {
     key = mrb_str_dup(mrb, key);
-    MRB_SET_FROZEN_FLAG(mrb_str_ptr(key));
+    mrb_str_ptr(key)->frozen = 1;
   }
   return key;
 }
@@ -1072,18 +1091,13 @@ h_replace(mrb_state *mrb, struct RHash *h, struct RHash *orig_h)
   }
 }
 
-void
+size_t
 mrb_gc_mark_hash(mrb_state *mrb, struct RHash *h)
 {
-  h_each(h, entry, {
+  H_EACH(h, entry) {
     mrb_gc_mark_value(mrb, entry->key);
     mrb_gc_mark_value(mrb, entry->val);
-  });
-}
-
-size_t
-mrb_gc_mark_hash_size(mrb_state *mrb, struct RHash *h)
-{
+  }
   return h_size(h) * 2;
 }
 
@@ -1108,9 +1122,22 @@ mrb_hash_memsize(mrb_value self)
 MRB_API void
 mrb_hash_foreach(mrb_state *mrb, struct RHash *h, mrb_hash_foreach_func *func, void *data)
 {
-  h_each(h, entry, {
-    if (func(mrb, entry->key, entry->val, data) != 0) return;
-  });
+  H_EACH(h, entry) {
+    int n;
+    H_CHECK_MODIFIED(mrb, h) {
+      n = func(mrb, entry->key, entry->val, data);
+    }
+    if (n != 0) return;
+  }
+}
+
+mrb_value
+mrb_hash_first_key(mrb_state *mrb, mrb_value h)
+{
+  H_EACH(mrb_hash_ptr(h), entry) {
+    return entry->key;
+  }
+  return mrb_nil_value();
 }
 
 MRB_API mrb_value
@@ -1247,6 +1274,21 @@ mrb_hash_set(mrb_state *mrb, mrb_value hash, mrb_value key, mrb_value val)
   mrb_field_write_barrier_value(mrb, mrb_basic_ptr(hash), key);
   mrb_field_write_barrier_value(mrb, mrb_basic_ptr(hash), val);
 }
+static void
+hash_set_default_proc(mrb_state *mrb, mrb_value hash, mrb_value proc)
+{
+  struct RProc *p = mrb_proc_ptr(proc);
+  if (MRB_PROC_STRICT_P(p)) {
+    int n = mrb_proc_arity(p);
+    if (n != 2 && (n >= 0 || n < -3)) {
+      if (n < 0) n = -n-1;
+      mrb_raisef(mrb, E_TYPE_ERROR, "default_proc takes two arguments (2 for %d)", n);
+    }
+  }
+  mrb_iv_set(mrb, hash, MRB_SYM(ifnone), proc);
+  RHASH(hash)->flags |= MRB_HASH_PROC_DEFAULT;
+  RHASH(hash)->flags |= MRB_HASH_DEFAULT;
+}
 
 /* 15.2.13.4.16 */
 /*
@@ -1297,10 +1339,10 @@ mrb_hash_init(mrb_state *mrb, mrb_value hash)
     if (ifnone_p) {
       mrb_argnum_error(mrb, 1, 0, 0);
     }
-    RHASH(hash)->flags |= MRB_HASH_PROC_DEFAULT;
-    ifnone = block;
+    hash_set_default_proc(mrb, hash, block);
+    return hash;
   }
-  if (!mrb_nil_p(ifnone)) {
+  if (ifnone_p && !mrb_nil_p(ifnone)) {
     RHASH(hash)->flags |= MRB_HASH_DEFAULT;
     mrb_iv_set(mrb, hash, MRB_SYM(ifnone), ifnone);
   }
@@ -1451,16 +1493,17 @@ mrb_hash_set_default_proc(mrb_state *mrb, mrb_value hash)
   mrb_value ifnone = mrb_get_arg1(mrb);
 
   hash_modify(mrb, hash);
+  if (!mrb_nil_p(ifnone)) {
+    mrb_check_type(mrb, ifnone, MRB_TT_PROC);
+  }
   mrb_iv_set(mrb, hash, MRB_SYM(ifnone), ifnone);
   if (!mrb_nil_p(ifnone)) {
-    RHASH(hash)->flags |= MRB_HASH_PROC_DEFAULT;
-    RHASH(hash)->flags |= MRB_HASH_DEFAULT;
+    hash_set_default_proc(mrb, hash, ifnone);
   }
   else {
     RHASH(hash)->flags &= ~MRB_HASH_DEFAULT;
     RHASH(hash)->flags &= ~MRB_HASH_PROC_DEFAULT;
   }
-
   return ifnone;
 }
 
@@ -1560,9 +1603,16 @@ mrb_hash_clear(mrb_state *mrb, mrb_value hash)
 static mrb_value
 mrb_hash_aset(mrb_state *mrb, mrb_value self)
 {
-  mrb_value key, val;
+  mrb_int argc = mrb_get_argc(mrb);
+
+  if (argc != 2) {
+    mrb_argnum_error(mrb, argc, 2, 2);
+  }
+
+  const mrb_value *argv = mrb_get_argv(mrb);
+  mrb_value key = argv[0];
+  mrb_value val = argv[1];
 
-  mrb_get_args(mrb, "oo", &key, &val);
   mrb_hash_set(mrb, self, key, val);
   return val;
 }
@@ -1634,9 +1684,9 @@ mrb_hash_keys(mrb_state *mrb, mrb_value hash)
 {
   struct RHash *h = mrb_hash_ptr(hash);
   mrb_value ary = mrb_ary_new_capa(mrb, (mrb_int)h_size(h));
-  h_each(h, entry, {
+  H_EACH(h, entry) {
     mrb_ary_push(mrb, ary, entry->key);
-  });
+  }
   return ary;
 }
 
@@ -1658,9 +1708,9 @@ mrb_hash_values(mrb_state *mrb, mrb_value hash)
 {
   struct RHash *h = mrb_hash_ptr(hash);
   mrb_value ary = mrb_ary_new_capa(mrb, (mrb_int)h_size(h));
-  h_each(h, entry, {
+  H_EACH(h, entry) {
     mrb_ary_push(mrb, ary, entry->val);
-  });
+  }
   return ary;
 }
 
@@ -1720,11 +1770,11 @@ mrb_hash_has_value(mrb_state *mrb, mrb_value hash)
 {
   mrb_value val = mrb_get_arg1(mrb);
   struct RHash *h = mrb_hash_ptr(hash);
-  h_each(h, entry, {
-    h_check_modified(mrb, h, {
+  H_EACH(h, entry) {
+    H_CHECK_MODIFIED(mrb, h) {
       if (mrb_equal(mrb, val, entry->val)) return mrb_true_value();
-    });
-  });
+    }
+  }
   return mrb_false_value();
 }
 
@@ -1740,11 +1790,11 @@ mrb_hash_merge(mrb_state *mrb, mrb_value hash1, mrb_value hash2)
 
   if (h1 == h2) return;
   if (h_size(h2) == 0) return;
-  h_each(h2, entry, {
-    h_check_modified(mrb, h2, {h_set(mrb, h1, entry->key, entry->val);});
-    mrb_field_write_barrier_value(mrb, (struct RBasic *)h1, entry->key);
-    mrb_field_write_barrier_value(mrb, (struct RBasic *)h1, entry->val);
-  });
+  H_EACH(h2, entry) {
+    H_CHECK_MODIFIED(mrb, h2) {h_set(mrb, h1, entry->key, entry->val);}
+    mrb_field_write_barrier_value(mrb, (struct RBasic*)h1, entry->key);
+    mrb_field_write_barrier_value(mrb, (struct RBasic*)h1, entry->val);
+  }
 }
 
 static mrb_value
@@ -1783,42 +1833,181 @@ mrb_hash_merge_m(mrb_state *mrb, mrb_value hash)
 static mrb_value
 mrb_hash_rehash(mrb_state *mrb, mrb_value self)
 {
+  hash_modify(mrb, self);
   h_rehash(mrb, mrb_hash_ptr(self));
   return self;
 }
 
+static mrb_value
+mrb_hash_compact(mrb_state *mrb, mrb_value hash)
+{
+  struct RHash *h = mrb_hash_ptr(hash);
+  mrb_bool ht_p = h_ht_p(h);
+  uint32_t size = ht_p ? ht_size(h) : ar_size(h);
+  uint32_t dec = 0;
+
+  hash_modify(mrb, hash);
+  H_EACH(h, entry) {
+    if (mrb_nil_p(entry->val)) {
+      entry_delete(entry);
+      dec++;
+    }
+  }
+  if (dec == 0) return mrb_nil_value();
+  size -= dec;
+  if (ht_p) {
+    ht_set_size(h, size);
+  }
+  else {
+    ar_set_size(h, size);
+  }
+  return hash;
+}
+
+/*
+ * call-seq:
+ *    hash.to_s    -> string
+ *    hash.inspect -> string
+ *
+ * Return the contents of this hash as a string.
+ */
+static mrb_value
+mrb_hash_to_s(mrb_state *mrb, mrb_value self)
+{
+  mrb->c->ci->mid = MRB_SYM(inspect);
+  mrb_value ret = mrb_str_new_lit(mrb, "{");
+  int ai = mrb_gc_arena_save(mrb);
+  if (mrb_inspect_recursive_p(mrb, self)) {
+    mrb_str_cat_lit(mrb, ret, "...}");
+    return ret;
+  }
+
+  mrb_int i = 0;
+  struct RHash *h = mrb_hash_ptr(self);
+  H_EACH(h, entry) {
+    if (i++ > 0) mrb_str_cat_lit(mrb, ret, ", ");
+    if (mrb_symbol_p(entry->key)) {
+      mrb_str_cat_str(mrb, ret, mrb_obj_as_string(mrb, entry->key));
+      mrb_gc_arena_restore(mrb, ai);
+      mrb_str_cat_lit(mrb, ret, ": ");
+    }
+    else {
+      H_CHECK_MODIFIED(mrb, h) {
+        mrb_str_cat_str(mrb, ret, mrb_inspect(mrb, entry->key));
+      }
+      mrb_gc_arena_restore(mrb, ai);
+      mrb_str_cat_lit(mrb, ret, " => ");
+    }
+    H_CHECK_MODIFIED(mrb, h) {
+      mrb_str_cat_str(mrb, ret, mrb_inspect(mrb, entry->val));
+    }
+    mrb_gc_arena_restore(mrb, ai);
+  }
+  mrb_str_cat_lit(mrb, ret, "}");
+
+  return ret;
+}
+
+/*
+ * call-seq:
+ *    hash.to_hash -> self
+ *
+ * Returns self.
+ */
+static mrb_value
+mrb_hash_to_hash(mrb_state *mrb, mrb_value self)
+{
+  return self;
+}
+
+/*
+ * call-seq:
+ *   hash.assoc(key) -> new_array or nil
+ *
+ * If the given key is found, returns a 2-element Array containing that key
+ * and its value:
+ *
+ *  h = {foo: 0, bar: 1, baz: 2}
+ *  h.assoc(:bar) # => [:bar, 1]
+ *
+ * Returns nil if key key is not found.
+ */
+static mrb_value
+mrb_hash_assoc(mrb_state *mrb, mrb_value hash)
+{
+  mrb_value key = mrb_get_arg1(mrb);
+  struct RHash *h = mrb_hash_ptr(hash);
+  H_EACH(h, entry) {
+    if (obj_eql(mrb, entry->key, key, h)) {
+      return mrb_assoc_new(mrb, entry->key, entry->val);
+    }
+  }
+  return mrb_nil_value();
+}
+
+/*
+ * call-seq:
+ *   hash.rassoc(value) -> new_array or nil
+ *
+ * Returns a new 2-element Array consisting of the key and value of the
+ * first-found entry whose value is == to value.
+ *
+ *  h = {foo: 0, bar: 1, baz: 1}
+ *  h.rassoc(1) # => [:bar, 1]
+ *
+ * Returns nil if no such value found.
+ */
+static mrb_value
+mrb_hash_rassoc(mrb_state *mrb, mrb_value hash)
+{
+  mrb_value value = mrb_get_arg1(mrb);
+  struct RHash *h = mrb_hash_ptr(hash);
+  H_EACH(h, entry) {
+    if (obj_eql(mrb, entry->val, value, h)) {
+      return mrb_assoc_new(mrb, entry->key, entry->val);
+    }
+  }
+  return mrb_nil_value();
+}
+
 void
 mrb_init_hash(mrb_state *mrb)
 {
   struct RClass *h;
 
-  mrb->hash_class = h = mrb_define_class(mrb, "Hash", mrb->object_class);              /* 15.2.13 */
+  mrb->hash_class = h = mrb_define_class_id(mrb, MRB_SYM(Hash), mrb->object_class);              /* 15.2.13 */
   MRB_SET_INSTANCE_TT(h, MRB_TT_HASH);
 
-  mrb_define_method(mrb, h, "[]",              mrb_hash_aget,        MRB_ARGS_REQ(1)); /* 15.2.13.4.2  */
-  mrb_define_method(mrb, h, "[]=",             mrb_hash_aset,        MRB_ARGS_REQ(2)); /* 15.2.13.4.3  */
-  mrb_define_method(mrb, h, "clear",           mrb_hash_clear,       MRB_ARGS_NONE()); /* 15.2.13.4.4  */
-  mrb_define_method(mrb, h, "default",         mrb_hash_default,     MRB_ARGS_OPT(1)); /* 15.2.13.4.5  */
-  mrb_define_method(mrb, h, "default=",        mrb_hash_set_default, MRB_ARGS_REQ(1)); /* 15.2.13.4.6  */
-  mrb_define_method(mrb, h, "default_proc",    mrb_hash_default_proc,MRB_ARGS_NONE()); /* 15.2.13.4.7  */
-  mrb_define_method(mrb, h, "default_proc=",   mrb_hash_set_default_proc,MRB_ARGS_REQ(1)); /* 15.2.13.4.7  */
-  mrb_define_method(mrb, h, "__delete",        mrb_hash_delete,      MRB_ARGS_REQ(1)); /* core of 15.2.13.4.8  */
-  mrb_define_method(mrb, h, "empty?",          mrb_hash_empty_m,     MRB_ARGS_NONE()); /* 15.2.13.4.12 */
-  mrb_define_method(mrb, h, "has_key?",        mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.13 */
-  mrb_define_method(mrb, h, "has_value?",      mrb_hash_has_value,   MRB_ARGS_REQ(1)); /* 15.2.13.4.14 */
-  mrb_define_method(mrb, h, "include?",        mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.15 */
-  mrb_define_method(mrb, h, "initialize",      mrb_hash_init,        MRB_ARGS_OPT(1)|MRB_ARGS_BLOCK()); /* 15.2.13.4.16 */
-  mrb_define_method(mrb, h, "initialize_copy", mrb_hash_init_copy,   MRB_ARGS_REQ(1)); /* 15.2.13.4.17 */
-  mrb_define_method(mrb, h, "key?",            mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.18 */
-  mrb_define_method(mrb, h, "keys",            mrb_hash_keys,        MRB_ARGS_NONE()); /* 15.2.13.4.19 */
-  mrb_define_method(mrb, h, "length",          mrb_hash_size_m,      MRB_ARGS_NONE()); /* 15.2.13.4.20 */
-  mrb_define_method(mrb, h, "member?",         mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.21 */
-  mrb_define_method(mrb, h, "replace",         mrb_hash_init_copy,   MRB_ARGS_REQ(1)); /* 15.2.13.4.23 */
-  mrb_define_method(mrb, h, "shift",           mrb_hash_shift,       MRB_ARGS_NONE()); /* 15.2.13.4.24 */
-  mrb_define_method(mrb, h, "size",            mrb_hash_size_m,      MRB_ARGS_NONE()); /* 15.2.13.4.25 */
-  mrb_define_method(mrb, h, "store",           mrb_hash_aset,        MRB_ARGS_REQ(2)); /* 15.2.13.4.26 */
-  mrb_define_method(mrb, h, "value?",          mrb_hash_has_value,   MRB_ARGS_REQ(1)); /* 15.2.13.4.27 */
-  mrb_define_method(mrb, h, "values",          mrb_hash_values,      MRB_ARGS_NONE()); /* 15.2.13.4.28 */
-  mrb_define_method(mrb, h, "rehash",          mrb_hash_rehash,      MRB_ARGS_NONE());
-  mrb_define_method(mrb, h, "__merge",         mrb_hash_merge_m,     MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, h, MRB_OPSYM(aref),          mrb_hash_aget,        MRB_ARGS_REQ(1)); /* 15.2.13.4.2  */
+  mrb_define_method_id(mrb, h, MRB_OPSYM(aset),          mrb_hash_aset,        MRB_ARGS_REQ(2)); /* 15.2.13.4.3  */
+  mrb_define_method_id(mrb, h, MRB_SYM(clear),           mrb_hash_clear,       MRB_ARGS_NONE()); /* 15.2.13.4.4  */
+  mrb_define_method_id(mrb, h, MRB_SYM(default),         mrb_hash_default,     MRB_ARGS_OPT(1)); /* 15.2.13.4.5  */
+  mrb_define_method_id(mrb, h, MRB_SYM_E(default),       mrb_hash_set_default, MRB_ARGS_REQ(1)); /* 15.2.13.4.6  */
+  mrb_define_method_id(mrb, h, MRB_SYM(default_proc),    mrb_hash_default_proc,MRB_ARGS_NONE()); /* 15.2.13.4.7  */
+  mrb_define_method_id(mrb, h, MRB_SYM_E(default_proc),  mrb_hash_set_default_proc,MRB_ARGS_REQ(1)); /* 15.2.13.4.7  */
+  mrb_define_method_id(mrb, h, MRB_SYM(__delete),        mrb_hash_delete,      MRB_ARGS_REQ(1)); /* core of 15.2.13.4.8  */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(empty),         mrb_hash_empty_m,     MRB_ARGS_NONE()); /* 15.2.13.4.12 */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(has_key),       mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.13 */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(has_value),     mrb_hash_has_value,   MRB_ARGS_REQ(1)); /* 15.2.13.4.14 */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(include),       mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.15 */
+  mrb_define_method_id(mrb, h, MRB_SYM(initialize),      mrb_hash_init,        MRB_ARGS_OPT(1)|MRB_ARGS_BLOCK()); /* 15.2.13.4.16 */
+  mrb_define_method_id(mrb, h, MRB_SYM(initialize_copy), mrb_hash_init_copy,   MRB_ARGS_REQ(1)); /* 15.2.13.4.17 */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(key),           mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.18 */
+  mrb_define_method_id(mrb, h, MRB_SYM(keys),            mrb_hash_keys,        MRB_ARGS_NONE()); /* 15.2.13.4.19 */
+  mrb_define_method_id(mrb, h, MRB_SYM(length),          mrb_hash_size_m,      MRB_ARGS_NONE()); /* 15.2.13.4.20 */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(member),        mrb_hash_has_key,     MRB_ARGS_REQ(1)); /* 15.2.13.4.21 */
+  mrb_define_method_id(mrb, h, MRB_SYM(replace),         mrb_hash_init_copy,   MRB_ARGS_REQ(1)); /* 15.2.13.4.23 */
+  mrb_define_method_id(mrb, h, MRB_SYM(shift),           mrb_hash_shift,       MRB_ARGS_NONE()); /* 15.2.13.4.24 */
+  mrb_define_method_id(mrb, h, MRB_SYM(size),            mrb_hash_size_m,      MRB_ARGS_NONE()); /* 15.2.13.4.25 */
+  mrb_define_method_id(mrb, h, MRB_SYM(store),           mrb_hash_aset,        MRB_ARGS_REQ(2)); /* 15.2.13.4.26 */
+  mrb_define_method_id(mrb, h, MRB_SYM_Q(value),         mrb_hash_has_value,   MRB_ARGS_REQ(1)); /* 15.2.13.4.27 */
+  mrb_define_method_id(mrb, h, MRB_SYM(values),          mrb_hash_values,      MRB_ARGS_NONE()); /* 15.2.13.4.28 */
+  mrb_define_method_id(mrb, h, MRB_SYM(to_s),            mrb_hash_to_s,        MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, h, MRB_SYM(inspect),         mrb_hash_to_s,        MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, h, MRB_SYM(rehash),          mrb_hash_rehash,      MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, h, MRB_SYM(to_hash),         mrb_hash_to_hash,     MRB_ARGS_NONE());
+  mrb_define_method_id(mrb, h, MRB_SYM(assoc),           mrb_hash_assoc,       MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, h, MRB_SYM(rassoc),          mrb_hash_rassoc,      MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, h, MRB_SYM(__merge),         mrb_hash_merge_m,     MRB_ARGS_REQ(1));
+  mrb_define_method_id(mrb, h, MRB_SYM(__compact),       mrb_hash_compact,     MRB_ARGS_NONE()); /* implementation of Hash#compact! */
 }
