@@ -19,11 +19,11 @@ def cmd_list(s)
 end
 
 def cmd(s)
-  return cmd_list(s).join(' ')
+  cmd_list(s).join(' ')
 end
 
 def cmd_bin(s)
-  return cmd_list(s).pop
+  cmd_list(s).pop
 end
 
 def shellquote(s)
@@ -44,7 +44,7 @@ ARGV.each do |gem|
 
   case RbConfig::CONFIG['host_os']
   when /mswin(?!ce)|mingw|bccwin/
-    gem = gem.gsub('\\', '/')
+    gem = gem.tr('\\', '/')
   end
 
   Dir["#{gem}/bintest/**/*.rb"].each do |file|
