@@ -67,7 +67,7 @@ assert('Module#append_features', '15.2.2.4.10') do
   end
 
   assert_equal Test4AppendFeatures2, Test4AppendFeatures2.const_get(:Const4AppendFeatures2)
-  assert_raise(FrozenError) { Module.new.append_features Class.new.freeze }
+  assert_raise(FrozenError) { Module.new.__send__(:append_features,Class.new.freeze) }
 end
 
 assert('Module#attr NameError') do
@@ -231,6 +231,18 @@ assert('Module#const_defined?', '15.2.2.4.20') do
   assert_true Test4ConstDefined.const_defined?(:Const4Test4ConstDefined)
   assert_false Test4ConstDefined.const_defined?(:NotExisting)
   assert_wrong_const_name{ Test4ConstDefined.const_defined?(:wrong_name) }
+
+  # shared empty iv_tbl (include)
+  m = Module.new
+  c = Class.new{include m}
+  m::CONST = 1
+  assert_true c.const_defined?(:CONST)
+
+  # shared empty iv_tbl (prepend)
+  m = Module.new
+  c = Class.new{prepend m}
+  m::CONST = 1
+  assert_true c.const_defined?(:CONST)
 end
 
 assert('Module#const_get', '15.2.2.4.21') do
@@ -246,6 +258,18 @@ assert('Module#const_get', '15.2.2.4.21') do
   assert_uninitialized_const{ Test4ConstGet.const_get(:I_DO_NOT_EXIST) }
   assert_uninitialized_const{ Test4ConstGet.const_get("I_DO_NOT_EXIST::ME_NEITHER") }
   assert_wrong_const_name{ Test4ConstGet.const_get(:wrong_name) }
+
+  # shared empty iv_tbl (include)
+  m = Module.new
+  c = Class.new{include m}
+  m::CONST = 1
+  assert_equal 1, c.const_get(:CONST)
+
+  # shared empty iv_tbl (prepend)
+  m = Module.new
+  c = Class.new{prepend m}
+  m::CONST = 1
+  assert_equal 1, c.const_get(:CONST)
 end
 
 assert('Module#const_set', '15.2.2.4.23') do
@@ -264,16 +288,15 @@ assert('Module#remove_const', '15.2.2.4.40') do
   module Test4RemoveConst
     ExistingConst = 23
   end
-
-  assert_equal 23, Test4RemoveConst.remove_const(:ExistingConst)
+  assert_equal 23, Test4RemoveConst.__send__(:remove_const,:ExistingConst)
   assert_false Test4RemoveConst.const_defined?(:ExistingConst)
   assert_raise_with_message_pattern(NameError, "constant * not defined") do
-    Test4RemoveConst.remove_const(:NonExistingConst)
+    Test4RemoveConst.__send__(:remove_const,:NonExistingConst)
   end
   %i[x X!].each do |n|
-    assert_wrong_const_name { Test4RemoveConst.remove_const(n) }
+    assert_wrong_const_name { Test4RemoveConst.__send__(:remove_const,n) }
   end
-  assert_raise(FrozenError) { Test4RemoveConst.freeze.remove_const(:A) }
+  assert_raise(FrozenError) { Test4RemoveConst.freeze.__send__(:remove_const,:A) }
 end
 
 assert('Module#const_missing', '15.2.2.4.22') do
@@ -291,11 +314,11 @@ assert('Module#extend_object', '15.2.2.4.25') do
   mod = Module.new { def foo; end }
   a = cls.new
   b = cls.new
-  mod.extend_object(b)
+  mod.__send__(:extend_object,b)
   assert_false a.respond_to?(:foo)
   assert_true b.respond_to?(:foo)
-  assert_raise(FrozenError) { mod.extend_object(cls.new.freeze) }
-  assert_raise(FrozenError, TypeError) { mod.extend_object(1) }
+  assert_raise(FrozenError) { mod.__send__(:extend_object,cls.new.freeze) }
+  assert_raise(FrozenError, TypeError) { mod.__send__(:extend_object,1) }
 end
 
 assert('Module#include', '15.2.2.4.27') do
@@ -356,16 +379,16 @@ end
 assert('Module#method_defined?', '15.2.2.4.34') do
   module Test4MethodDefined
     module A
-      def method1()  end
+      def method1() end
     end
 
     class B
-      def method2()  end
+      def method2() end
     end
 
     class C < B
       include A
-      def method3()  end
+      def method3() end
     end
   end
 
@@ -449,9 +472,9 @@ assert 'Module#prepend_features' do
   mod = Module.new { def m; :mod end }
   cls = Class.new { def m; :cls end }
   assert_equal :cls, cls.new.m
-  mod.prepend_features(cls)
+  mod.__send__(:prepend_features,cls)
   assert_equal :mod, cls.new.m
-  assert_raise(FrozenError) { Module.new.prepend_features(Class.new.freeze) }
+  assert_raise(FrozenError) { Module.new.__send__(:prepend_features,Class.new.freeze) }
 end
 
 # @!group prepend
@@ -542,8 +565,8 @@ end
 
     bug6662 = '[ruby-dev:45868]'
     c2 = labeled_class("c2", c)
-    anc = c2.ancestors
-    assert_equal([c2, m, c, Object], anc[0..anc.index(Object)], bug6662)
+    as = c2.ancestors
+    assert_equal([c2, m, c, Object], as[0..as.index(Object)], bug6662)
   end
 
   assert 'Module#prepend + Module#ancestors' do
@@ -686,7 +709,7 @@ end
   #    assert_nothing_raised(SystemStackError, bug10847) do
   #      0.3.numerator
   #    end
-  #  end;
+  #  end
   #end
 
   assert 'Module#prepend to frozen class' do
@@ -729,14 +752,14 @@ end
 
 assert('Issue 1467') do
   module M1
-    def initialize()
+    def initialize
       super()
     end
   end
 
   class C1
     include M1
-      def initialize()
+      def initialize
         super()
       end
   end
@@ -763,6 +786,47 @@ assert('clone Module') do
   assert_true(B.new.foo)
 end
 
+assert('method visibility') do
+  class CallTypeTest
+    def test_private(&block)
+      func(&block)
+    end
+    def test_protected(&block)
+      self.func(&block)
+    end
+    private
+    def func
+      yield
+    end
+  end
+
+  v = CallTypeTest.new
+
+  assert_raise(NoMethodError) { v.func { :test } }
+  assert_equal :test, v.test_private { :test }
+
+  class CallTypeTest
+    protected :func
+  end
+
+  assert_raise(NoMethodError) { v.func { :test } }
+  assert_equal :test, v.test_protected { :test }
+  assert_equal :test, v.test_private { :test }
+
+  class CallTypeTest
+    public def public_func
+      :test
+    end
+
+    public :func
+  end
+
+  assert_equal :test, v.public_func
+  assert_equal :test, v.func { :test }
+  assert_equal :test, v.test_protected { :test }
+  assert_equal :test, v.test_private { :test }
+end
+
 assert('Module#module_function') do
   module M
     def modfunc; end
@@ -779,7 +843,7 @@ end
 
 assert('module to return the last value') do
   m = module M; :m end
-  assert_equal(m, :m)
+  assert_equal(:m, m)
 end
 
 assert('module to return nil if body is empty') do
@@ -796,3 +860,25 @@ assert('get constant of parent module in singleton class; issue #3568') do
 
   assert_equal("value", actual)
 end
+
+assert('shared empty iv_tbl (include)') do
+  m1 = Module.new
+  m2 = Module.new{include m1}
+  c = Class.new{include m2}
+  m1::CONST1 = 1
+  assert_equal 1, m2::CONST1
+  assert_equal 1, c::CONST1
+  m2::CONST2 = 2
+  assert_equal 2, c::CONST2
+end
+
+assert('shared empty iv_tbl (prepend)') do
+  m1 = Module.new
+  m2 = Module.new{prepend m1}
+  c = Class.new{include m2}
+  m1::CONST1 = 1
+  assert_equal 1, m2::CONST1
+  assert_equal 1, c::CONST1
+  m2::CONST2 = 2
+  assert_equal 2, c::CONST2
+end
